"""
These tests are based on the API calls made to hdate by homeassistant (and
maybe other apps in the future).
"""

import sys
from datetime import date, datetime

from hdate import HDate, Location, Zmanim

_ASTRAL = "astral" in sys.modules


class TestHDateAPI:
    """Test the HDate API provided in the README."""

    def test_readme_example_english(self, capsys):
        """Test the README example in English."""

        test_date = date(2016, 4, 18)
        hdate = HDate(test_date, hebrew=False)
        print(hdate)
        captured = capsys.readouterr()
        assert captured.out == "Monday 10 Nisan 5776\n"

    def test_readme_example_hebrew(self, capsys):
        """Test the README example in Hebrew."""
        test_date = date(2016, 4, 26)
        hdate = HDate(test_date, hebrew=True)
        print(hdate)
        captured = capsys.readouterr()
        assert captured.out == "יום שלישי י\"ח בניסן ה' תשע\"ו ג' בעומר חול המועד פסח\n"

    def test_get_hebrew_date(self):
        """Print the hebrew date."""
        test_date = datetime(2018, 11, 2)
        assert HDate(test_date).hebrew_date == 'כ"ד מרחשוון ה\' תשע"ט'
        assert HDate(test_date, hebrew=False).hebrew_date == "24 Marcheshvan 5779"

    def test_get_upcoming_parasha(self):
        """Check that the upcoming parasha is correct."""
        test_date = datetime(2018, 11, 2)
        assert HDate(test_date).parasha == "חיי שרה"
        assert HDate(test_date, hebrew=False).parasha == "Chayei Sara"

    def test_get_upcoming_parasha_vezot_habracha(self):
        """Check that the upcoming parasha is correct for vezot habracha."""
        test_date = datetime(2018, 9, 30)
        assert HDate(test_date).parasha == "וזאת הברכה"
        assert HDate(test_date, hebrew=False).parasha == "Vezot Habracha"

    def test_get_holiday_description(self):
        """Check that the holiday description is correct."""
        test_date = datetime(2018, 12, 3)
        assert HDate(test_date).holiday_description == "חנוכה"
        assert HDate(test_date, hebrew=False).holiday_description == "Chanukah"


class TestZmanimAPI:
    """Test the API provided in the README."""

    def test_readme_example_hebrew(self, capsys):
        """Test for hebrew."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=date(2016, 4, 18), location=coord)
        print(zman)
        captured = capsys.readouterr()
        if not _ASTRAL:
            return
        assert (
            captured.out == "עלות השחר - 04:52:00\n"
            "זמן טלית ותפילין - 05:18:00\n"
            "הנץ החמה - 06:08:00\n"
            'סוף זמן ק"ש מג"א - 08:46:00\n'
            'סוף זמן ק"ש גר"א - 09:23:00\n'
            'סוף זמן תפילה מג"א - 10:04:00\n'
            'סוף זמן תפילה גר"א - 10:28:00\n'
            "חצות היום - 12:40:00\n"
            "מנחה גדולה - 13:10:30\n"
            "מנחה קטנה - 16:25:30\n"
            "פלג המנחה - 17:50:45\n"
            "שקיעה - 19:12:00\n"
            "צאת הכוכבים - 19:38:00\n"
            "חצות הלילה - 00:40:00\n"
        )

    def test_readme_example_english(self, capsys):
        """Test for english."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=date(2016, 4, 18), location=coord, hebrew=False)
        print(zman)
        captured = capsys.readouterr()
        if not _ASTRAL:
            return
        assert (
            captured.out == "Alot HaShachar - 04:52:00\n"
            "Talit & Tefilin's time - 05:18:00\n"
            "Sunrise - 06:08:00\n"
            'Shema EOT MG"A - 08:46:00\n'
            'Shema EOT GR"A - 09:23:00\n'
            'Tefila EOT MG"A - 10:04:00\n'
            'Tefila EOT GR"A - 10:28:00\n'
            "Midday - 12:40:00\n"
            "Big Mincha - 13:10:30\n"
            "Small Mincha - 16:25:30\n"
            "Plag Mincha - 17:50:45\n"
            "Sunset - 19:12:00\n"
            "First stars - 19:38:00\n"
            "Midnight - 00:40:00\n"
        )

    def test_issur_melacha_weekday(self):
        """Test for issur melacha on a weekday."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=date(2018, 11, 12), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_issur_melacha_shabbat_morning(self):
        """Test for issur melacha on shabbat morning."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2018, 11, 10, 9), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_friday_morning(self):
        """Test for issur melacha on friday morning."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2018, 11, 9, 9, 45), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_issur_melacha_friday_evening(self):
        """Test for issur melacha on friday evening."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2018, 11, 9, 16, 45), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_motsaei_shabbat(self):
        """Test for issur melacha on Motsaei shabbat."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2018, 11, 10, 17, 45), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_issur_melacha_shavuot_morning(self):
        """Test for issur melacha on shavuot morning."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2019, 6, 9, 9), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_pesach_vi_morning(self):
        """Test for issur melacha on erev shvii shel pesach morning."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2019, 4, 25, 9, 45), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_issur_melacha_shavuot_evening(self):
        """Test for issur melacha on shavuot evening."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2019, 6, 8, 21, 45), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_motsaei_shavuot(self):
        """Test for issur melacha on motsaei shavuot."""
        coord = Location("פתח תקוה", 32.08707, 34.88747, "Asia/Jerusalem", 54)
        zman = Zmanim(date=datetime(2019, 6, 9, 20, 30), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_issur_melacha_pesach_ii_morning(self):
        """Test for issur melacha on the second day of pesach in the diaspora."""
        coord = Location(
            name="New York",
            latitude=40.7128,
            longitude=-74.0060,
            timezone="America/New_York",
            diaspora=True,
        )
        zman = Zmanim(date=datetime(2019, 4, 21, 9), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_pesach_ii_evening(self):
        """Test for issur melacha on the eve of second day of pesach in the diaspora."""
        coord = Location(
            name="New York",
            latitude=40.7128,
            longitude=-74.0060,
            timezone="America/New_York",
            diaspora=True,
        )
        zman = Zmanim(date=datetime(2019, 4, 20, 21, 45), location=coord)
        assert zman.issur_melacha_in_effect

    def test_issur_melacha_motsaei_pesach_ii(self):
        """Test for issur melacha on the end of second day of pesach in the diaspora."""
        coord = Location(
            name="New York",
            latitude=40.7128,
            longitude=-74.0060,
            timezone="America/New_York",
            diaspora=True,
        )
        zman = Zmanim(date=datetime(2019, 4, 21, 20, 30), location=coord)
        assert not zman.issur_melacha_in_effect

    def test_zmanim_localized_datetime(self):
        """Test for issur melacha if datetime is localized."""
        coord = Location(
            name="New York",
            latitude=40.7128,
            longitude=-74.0060,
            timezone="America/New_York",
            diaspora=True,
        )
        zman = Zmanim(
            date=datetime(2019, 4, 21, 20, 30, tzinfo=coord.timezone), location=coord
        )
        assert not zman.issur_melacha_in_effect



"""Pure Hebrew date class."""

from dataclasses import dataclass

from hdate.htables import Months


@dataclass
class HebrewDate:
    """Define a Hebrew date object."""

    year: int
    month: Months
    day: int

    def __post_init__(self):
        self.month = (
            self.month if isinstance(self.month, Months) else Months(self.month)
        )



"""Test Zmanim objects."""

import datetime
import random
import sys
from calendar import isleap
from datetime import datetime as dt
from datetime import timedelta as td

import pytest

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from hdate import Zmanim
from hdate.location import Location

_ASTRAL = "astral" in sys.modules

NYC_LAT = 40.7128
NYC_LNG = -74.0060

LONDON_LAT = 51.5074
LONDON_LNG = -0.1278

PUNTA_ARENAS_LAT = -53.1678  # Southern example
PUNTA_ARENAS_LNG = -70.9167


def compare_dates(date1, date2, allow_grace=False):
    """Compare 2 dates to be more or less equal."""
    if not (date1 or date2):
        assert date1 == date2
    else:
        grace = td(minutes=5 if (not _ASTRAL or allow_grace) else 0)
        assert date1 - grace <= date2 <= date1 + grace


def compare_times(time1, time2, allow_grace=False):
    """Compare times to be equal."""
    compare_dates(
        dt.combine(dt.today(), time1), dt.combine(dt.today(), time2), allow_grace
    )


class TestZmanim:
    """Zmanim tests"""

    def test_bad_date(self):
        """Check that a bad value argument to zmanim raises an error"""
        with pytest.raises(TypeError):
            Zmanim(date="bad value")

    @pytest.mark.parametrize("execution_number", list(range(5)))
    def test_same_doy_is_equal(self, execution_number, random_date):
        """Test two doy to be equal."""
        print(f"Run number {execution_number}")
        other_year = random.randint(500, 3000)
        shift_day = datetime.timedelta(days=0)
        this_date = random_date

        if isleap(this_date.year) != isleap(other_year) and this_date > datetime.date(
            this_date.year, 2, 28
        ):
            if isleap(other_year):
                shift_day = datetime.timedelta(days=-1)
            else:
                shift_day = datetime.timedelta(days=1)

        if (
            isleap(this_date.year)
            and not isleap(other_year)
            and this_date.month == 2
            and this_date.day == 29
        ):
            # Special case we can't simply replace the year as there's
            # no leap day in the other year
            other_date = datetime.date(other_year, 3, 1)
        else:
            other_date = this_date.replace(year=other_year) + shift_day

        this_zmanim = Zmanim(this_date).get_utc_sun_time_full()
        other_zmanim = Zmanim(other_date).get_utc_sun_time_full()
        grace = 0 if not _ASTRAL else 14
        for key, value in this_zmanim.items():
            assert value - grace <= other_zmanim[key] <= value + grace, key

    def test_extreme_zmanim(self):
        """Test that Zmanim north to 50 degrees latitude is correct."""
        day = datetime.date(2024, 6, 18)
        compare_times(
            Zmanim(
                date=day,
                location=Location(
                    name="London",
                    latitude=LONDON_LAT,
                    longitude=LONDON_LNG,
                    timezone="Europe/London",
                    diaspora=True,
                ),
            )
            .zmanim["sunset"]
            .time(),
            datetime.time(21, 22),
            allow_grace=True,
        )
        compare_times(
            Zmanim(
                date=day,
                location=Location(
                    name="Punta Arenas",
                    latitude=PUNTA_ARENAS_LAT,
                    longitude=PUNTA_ARENAS_LNG,
                    timezone="America/Punta_Arenas",
                    diaspora=True,
                ),
            )
            .zmanim["sunset"]
            .time(),
            datetime.time(17, 31),
            allow_grace=True,
        )

    def test_using_tzinfo(self):
        """Test tzinfo to be correct."""
        day = datetime.date(2018, 9, 8)
        timezone_str = "America/New_York"
        timezone = ZoneInfo(timezone_str)
        location_tz_str = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone=timezone_str,
            diaspora=True,
        )
        location = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone=timezone,
            diaspora=True,
        )

        compare_times(
            Zmanim(date=day, location=location_tz_str).zmanim["first_stars"].time(),
            datetime.time(19, 45),
        )

        compare_times(
            Zmanim(date=day, location=location).zmanim["first_stars"].time(),
            datetime.time(19, 45),
        )

    # Times are assumed for NYC.
    CANDLES_TEST = [
        (dt(2018, 9, 7, 13, 1), 18, dt(2018, 9, 7, 19, 0), False),
        (dt(2018, 9, 7, 19, 4), 18, dt(2018, 9, 7, 19, 0), True),
        (dt(2018, 9, 8, 13, 1), 18, None, True),
        (dt(2018, 9, 19, 22, 1), 18, None, False),
        (dt(2018, 9, 9, 16, 1), 20, dt(2018, 9, 9, 18, 55), False),
        (dt(2018, 9, 9, 19, 30), 18, dt(2018, 9, 9, 18, 57), True),
        # Candle lighting matches the time that would be havdalah.
        (dt(2018, 9, 10, 8, 1), 18, dt(2018, 9, 10, 19, 55), True),
        (dt(2018, 9, 10, 20, 20), 18, dt(2018, 9, 10, 19, 55), True),
    ]

    @pytest.mark.parametrize(
        ["now", "offset", "candle_lighting", "melacha_assur"], CANDLES_TEST
    )
    def test_candle_lighting(self, now, offset, candle_lighting, melacha_assur):
        """Test candle lighting values."""
        location_tz_str = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone="America/New_York",
            diaspora=True,
        )
        # Use a constant offset for Havdalah for unit test stability.
        zmanim = Zmanim(
            date=now,
            location=location_tz_str,
            candle_lighting_offset=offset,
            havdalah_offset=42,
        )
        actual = zmanim.candle_lighting
        if actual is not None:
            actual = actual.replace(tzinfo=None)
        compare_dates(actual, candle_lighting)
        assert zmanim.issur_melacha_in_effect == melacha_assur

    # Times are assumed for NYC.
    HAVDALAH_TEST = [
        (dt(2018, 9, 7, 13, 1), 42, None, False),
        (dt(2018, 9, 7, 20, 1), 42, None, True),
        (dt(2018, 9, 8, 13, 1), 42, dt(2018, 9, 8, 19, 59), True),
        (dt(2018, 9, 8, 13, 1), 0, dt(2018, 9, 8, 19, 58), True),
        (dt(2018, 9, 19, 22, 1), 18, dt(2018, 9, 19, 19, 16), False),
        (dt(2018, 9, 9, 16, 1), 0, None, False),
        (dt(2018, 9, 9, 19, 30), 0, None, True),
        (dt(2018, 9, 11, 16, 1), 0, dt(2018, 9, 11, 19, 53), True),
        # No havdalah in the middle of Yom Tov.
        (dt(2018, 9, 10, 8, 1), 0, None, True),
        (dt(2018, 9, 10, 20, 20), 0, None, True),
    ]

    @pytest.mark.parametrize(
        ["now", "offset", "havdalah", "melacha_assur"], HAVDALAH_TEST
    )
    def test_havdalah(self, now, offset, havdalah, melacha_assur):
        """Test havdalah times."""
        location_tz_str = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone="America/New_York",
            diaspora=True,
        )
        # Use a constant offset for Havdalah for unit test stability.
        zmanim = Zmanim(date=now, location=location_tz_str, havdalah_offset=offset)
        actual = zmanim.havdalah
        if actual is not None:
            actual = actual.replace(tzinfo=None)
        compare_dates(actual, havdalah)
        assert zmanim.issur_melacha_in_effect == melacha_assur

    # Times are assumed for NYC.
    MOTZEI_SHABBAT_CHAG_TEST = [
        (dt(2018, 9, 7, 13, 1), 42, False),
        (dt(2018, 9, 7, 20, 1), 42, False),
        (dt(2018, 9, 8, 13, 1), 42, False),
        (dt(2018, 9, 8, 22, 1), 0, True),
        (dt(2018, 9, 9, 16, 1), 0, False),
        (dt(2018, 9, 9, 19, 30), 0, False),
        (dt(2018, 9, 10, 16, 1), 0, False),
        (dt(2018, 9, 10, 19, 30), 0, False),
        (dt(2018, 9, 11, 16, 1), 0, False),
        (dt(2018, 9, 11, 20, 1), 0, True),
        (dt(2018, 9, 19, 22, 1), 18, True),
    ]

    @pytest.mark.parametrize(
        ["now", "offset", "motzei_shabbat_chag"], MOTZEI_SHABBAT_CHAG_TEST
    )
    def test_motzei_shabbat_chag(self, now, offset, motzei_shabbat_chag):
        """Test motzei shabbat chag boolean is correct."""
        location_tz_str = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone="America/New_York",
            diaspora=True,
        )
        # Use a constant offset for Havdalah for unit test stability.
        zmanim = Zmanim(date=now, location=location_tz_str, havdalah_offset=offset)
        assert zmanim.motzei_shabbat_chag == motzei_shabbat_chag

    # Times are assumed for NYC.
    EREV_SHABBAT_CHAG_TEST = [
        (dt(2018, 9, 7, 13, 1), 42, True),  # fri
        (dt(2018, 9, 7, 20, 1), 42, False),
        (dt(2018, 9, 8, 13, 1), 42, False),  # sat
        (dt(2018, 9, 8, 20, 1), 0, False),
        (dt(2018, 9, 9, 16, 1), 0, True),  # erev rosh hashana
        (dt(2018, 9, 9, 19, 30), 0, False),
        (dt(2018, 9, 10, 16, 1), 0, False),  # rosh hashana I
        (dt(2018, 9, 10, 19, 30), 0, False),
        (dt(2018, 9, 11, 16, 1), 0, False),  # rosh hashana II
        (dt(2018, 9, 11, 20, 1), 0, False),
        (dt(2018, 9, 18, 13, 1), 18, True),  # erev yom kipur
        (dt(2018, 9, 18, 22, 1), 18, False),
        (dt(2018, 9, 19, 13, 1), 18, False),
        (dt(2018, 9, 19, 22, 1), 18, False),
    ]

    @pytest.mark.parametrize(
        ["now", "offset", "erev_shabbat_chag"], EREV_SHABBAT_CHAG_TEST
    )
    def test_erev_shabbat_hag(self, now, offset, erev_shabbat_chag):
        """Test erev shabbat chag boolean is correct."""
        location_tz_str = Location(
            name="New York",
            latitude=NYC_LAT,
            longitude=NYC_LNG,
            timezone="America/New_York",
            diaspora=True,
        )
        # Use a constant offset for Havdalah for unit test stability.
        zmanim = Zmanim(date=now, location=location_tz_str, havdalah_offset=offset)
        assert zmanim.erev_shabbat_chag == erev_shabbat_chag



"""
Jewish calendrical times for a given location.

HDate calculates and generates a representation either in English or Hebrew
of the Jewish calendrical times for a given location
"""

import datetime as dt
import logging
import math

from hdate import htables
from hdate.date import HDate
from hdate.location import Location

try:
    import astral
    import astral.sun

    _USE_ASTRAL = True
except ImportError:
    _USE_ASTRAL = False

MAX_LATITUDE_ASTRAL = 50.0
_LOGGER = logging.getLogger(__name__)


class Zmanim:  # pylint: disable=too-many-instance-attributes
    """Return Jewish day times.

    The Zmanim class returns times for the specified day ONLY. If you wish to
    obtain times for the interval of a multi-day holiday for example, you need
    to use Zmanim in conjunction with some of the iterative properties of
    HDate. Also, Zmanim are reported regardless of the current time. So the
    havdalah value is constant if the current time is before or after it.
    The current time is only used to report the "issur_melacha_in_effect"
    property.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        date=dt.datetime.now(),
        location=Location(),
        hebrew=True,
        candle_lighting_offset=18,
        havdalah_offset=0,
    ):
        """
        Initialize the Zmanim object.

        As the timezone is expected to be part of the location object, any
        tzinfo passed along is discarded. Essentially making the datetime
        object non-timezone aware.

        The time zone information is appended to the date received based on the
        location object. After which it is transformed to UTC for all internal
        calculations.
        """
        self.location = location
        self.hebrew = hebrew
        self.candle_lighting_offset = candle_lighting_offset
        self.havdalah_offset = havdalah_offset

        # If a non-timezone aware date is received, use timezone from location
        # to make it timezone aware and change to UTC for calculations.

        # If timezone aware is received as date, we expect it to match the
        # timezone specified by location, so it can be overridden and changed
        # to UTC for calculations as above.
        if isinstance(date, dt.datetime):
            _LOGGER.debug("Date input is of type datetime: %r", date)
            self.date = date.date()
            self.time = date.replace(tzinfo=None)
        elif isinstance(date, dt.date):
            _LOGGER.debug("Date input is of type date: %r", date)
            self.date = date
            self.time = dt.datetime.now()
        else:
            raise TypeError

        _LOGGER.debug("Resetting timezone to UTC for calculations")
        self.time = self.time.replace(tzinfo=location.timezone).astimezone(
            dt.timezone.utc
        )

        if _USE_ASTRAL and (abs(location.latitude) <= MAX_LATITUDE_ASTRAL):
            self.astral_observer = astral.Observer(
                latitude=self.location.latitude, longitude=self.location.longitude
            )
            self.astral_sun = astral.sun.sun(self.astral_observer, self.date)

    def __str__(self):
        """Return a Unicode representation of Zmanim."""
        return "\n".join(
            [
                f"{zman.description[self.hebrew + 1]} - {self.zmanim[zman.zman].time()}"
                for zman in htables.ZMANIM
            ]
        )

    def __repr__(self):
        """Return a representation of Zmanim for programmatic use."""
        # As time zone information is not really reusable due to DST, when
        # creating a __repr__ of zmanim, we show a timezone naive datetime.
        return (
            "Zmanim(date="
            f"{self.time.astimezone(self.location.timezone).replace(tzinfo=None)!r},"
            f" location={self.location!r}, hebrew={self.hebrew})"
        )

    @property
    def utc_zmanim(self):
        """Return a dictionary of the zmanim in UTC time format."""
        basetime = dt.datetime.combine(self.date, dt.time()).replace(
            tzinfo=dt.timezone.utc
        )
        _LOGGER.debug("Calculating UTC zmanim for %r", basetime)
        return {
            key: basetime + dt.timedelta(minutes=value)
            for key, value in self.get_utc_sun_time_full().items()
        }

    @property
    def zmanim(self):
        """Return a dictionary of the zmanim the object represents."""
        return {
            key: value.astimezone(self.location.timezone)
            for key, value in self.utc_zmanim.items()
        }

    @property
    def candle_lighting(self):
        """Return the time for candle lighting, or None if not applicable."""
        today = HDate(gdate=self.date, diaspora=self.location.diaspora)
        tomorrow = HDate(
            gdate=self.date + dt.timedelta(days=1), diaspora=self.location.diaspora
        )

        # If today is a Yom Tov or Shabbat, and tomorrow is a Yom Tov or
        # Shabbat return the havdalah time as the candle lighting time.
        if (today.is_yom_tov or today.is_shabbat) and (
            tomorrow.is_yom_tov or tomorrow.is_shabbat
        ):
            return self._havdalah_datetime

        # Otherwise, if today is Friday or erev Yom Tov, return candle
        # lighting.
        if tomorrow.is_shabbat or tomorrow.is_yom_tov:
            return self.zmanim["sunset"] - dt.timedelta(
                minutes=self.candle_lighting_offset
            )
        return None

    @property
    def _havdalah_datetime(self):
        """Compute the havdalah time based on settings."""
        if self.havdalah_offset == 0:
            return self.zmanim["three_stars"]
        # Otherwise, use the offset.
        return self.zmanim["sunset"] + dt.timedelta(minutes=self.havdalah_offset)

    @property
    def havdalah(self):
        """Return the time for havdalah, or None if not applicable.

        If havdalah_offset is 0, uses the time for three_stars. Otherwise,
        adds the offset to the time of sunset and uses that.
        If it's currently a multi-day YomTov, and the end of the stretch is
        after today, the havdalah value is defined to be None (to avoid
        misleading the user that melacha is permitted).
        """
        today = HDate(gdate=self.date, diaspora=self.location.diaspora)
        tomorrow = HDate(
            gdate=self.date + dt.timedelta(days=1), diaspora=self.location.diaspora
        )

        # If today is Yom Tov or Shabbat, and tomorrow is Yom Tov or Shabbat,
        # then there is no havdalah value for today. Technically, there is
        # havdalah mikodesh l'kodesh, but that is represented in the
        # candle_lighting value to avoid misuse of the havdalah API.
        if today.is_shabbat or today.is_yom_tov:
            if tomorrow.is_shabbat or tomorrow.is_yom_tov:
                return None
            return self._havdalah_datetime
        return None

    @property
    def issur_melacha_in_effect(self):
        """At the given time, return whether issur melacha is in effect."""
        today = HDate(gdate=self.date, diaspora=self.location.diaspora)
        tomorrow = HDate(
            gdate=self.date + dt.timedelta(days=1), diaspora=self.location.diaspora
        )

        if (today.is_shabbat or today.is_yom_tov) and (
            tomorrow.is_shabbat or tomorrow.is_yom_tov
        ):
            return True
        if (today.is_shabbat or today.is_yom_tov) and (self.time < self.havdalah):
            return True
        if (tomorrow.is_shabbat or tomorrow.is_yom_tov) and (
            self.time >= self.candle_lighting
        ):
            return True

        return False

    @property
    def erev_shabbat_chag(self):
        """At the given time, return whether erev shabbat or chag"""
        today = HDate(gdate=self.date, diaspora=self.location.diaspora)
        tomorrow = HDate(
            gdate=self.date + dt.timedelta(days=1), diaspora=self.location.diaspora
        )

        if (
            (tomorrow.is_shabbat or tomorrow.is_yom_tov)
            and (not today.is_shabbat and not today.is_yom_tov)
            and (self.time < self.candle_lighting)
        ):
            return True

        return False

    @property
    def motzei_shabbat_chag(self):
        """At the given time, return whether motzei shabbat or chag"""
        today = HDate(gdate=self.date, diaspora=self.location.diaspora)
        tomorrow = HDate(
            gdate=self.date + dt.timedelta(days=1), diaspora=self.location.diaspora
        )

        if (today.is_shabbat or today.is_yom_tov) and (
            tomorrow.is_shabbat or tomorrow.is_yom_tov
        ):
            return False
        if (today.is_shabbat or today.is_yom_tov) and (self.time > self.havdalah):
            return True

        return False

    def gday_of_year(self):
        """Return the number of days since January 1 of the given year."""
        return (self.date - dt.date(self.date.year, 1, 1)).days

    def _get_utc_sun_time_deg(self, deg):
        """
        Return the times in minutes from 00:00 (utc) for a given sun altitude.

        This is done for a given sun altitude in sunrise `deg` degrees
        This function only works for altitudes sun really is.
        If the sun never gets to this altitude, the returned sunset and sunrise
        values will be negative. This can happen in low altitude when latitude
        is nearing the poles in winter times, the sun never goes very high in
        the sky there.

        Algorithm from
        http://www.srrb.noaa.gov/highlights/sunrise/calcdetails.html
        The low accuracy solar position equations are used.
        These routines are based on Jean Meeus's book Astronomical Algorithms.
        """
        gama = 0  # location of sun in yearly cycle in radians
        eqtime = 0  # difference betwen sun noon and clock noon
        decl = 0  # sun declanation
        hour_angle = 0  # solar hour angle
        sunrise_angle = math.pi * deg / 180.0  # sun angle at sunrise/set

        # get the day of year
        day_of_year = self.gday_of_year()

        # get radians of sun orbit around earth =)
        gama = 2.0 * math.pi * ((day_of_year - 1) / 365.0)

        # get the diff betwen suns clock and wall clock in minutes
        eqtime = 229.18 * (
            0.000075
            + 0.001868 * math.cos(gama)
            - 0.032077 * math.sin(gama)
            - 0.014615 * math.cos(2.0 * gama)
            - 0.040849 * math.sin(2.0 * gama)
        )

        # calculate suns declanation at the equater in radians
        decl = (
            0.006918
            - 0.399912 * math.cos(gama)
            + 0.070257 * math.sin(gama)
            - 0.006758 * math.cos(2.0 * gama)
            + 0.000907 * math.sin(2.0 * gama)
            - 0.002697 * math.cos(3.0 * gama)
            + 0.00148 * math.sin(3.0 * gama)
        )

        # we use radians, ratio is 2pi/360
        latitude = math.pi * self.location.latitude / 180.0

        # the sun real time diff from noon at sunset/rise in radians
        try:
            hour_angle = math.acos(
                math.cos(sunrise_angle) / (math.cos(latitude) * math.cos(decl))
                - math.tan(latitude) * math.tan(decl)
            )
        # check for too high altitudes and return negative values
        except ValueError:
            return -720, -720

        # we use minutes, ratio is 1440min/2pi
        hour_angle = 720.0 * hour_angle / math.pi

        # get sunset/rise times in utc wall clock in minutes from 00:00 time
        # sunrise / sunset
        longitude = self.location.longitude
        return (
            int(720.0 - 4.0 * longitude - hour_angle - eqtime),
            int(720.0 - 4.0 * longitude + hour_angle - eqtime),
        )

    def _datetime_to_minutes_offest(self, time):
        """Return the time in minutes from 00:00 (utc) for a given time."""
        return (
            time.hour * 60
            + time.minute
            + (1 if time.second >= 30 else 0)
            + int((time.date() - self.date).total_seconds() // 60)
        )

    def _get_utc_time_of_transit(self, zenith, rising):
        """Return the time in minutes from 00:00 (utc) for a given sun altitude."""
        return self._datetime_to_minutes_offest(
            astral.sun.time_of_transit(
                self.astral_observer,
                self.date,
                zenith,
                astral.SunDirection.RISING if rising else astral.SunDirection.SETTING,
            )
        )

    def get_utc_sun_time_full(self):
        """Return a list of Jewish times for the given location."""
        if (not _USE_ASTRAL) or (abs(self.location.latitude) > MAX_LATITUDE_ASTRAL):
            sunrise, sunset = self._get_utc_sun_time_deg(90.833)
            first_light, _ = self._get_utc_sun_time_deg(106.1)
            talit, _ = self._get_utc_sun_time_deg(101.0)
            _, first_stars = self._get_utc_sun_time_deg(96.0)
            _, three_stars = self._get_utc_sun_time_deg(98.5)
        else:
            sunrise = self._datetime_to_minutes_offest(self.astral_sun["sunrise"])
            sunset = self._datetime_to_minutes_offest(self.astral_sun["sunset"])
            first_light = self._get_utc_time_of_transit(106.1, True)
            talit = self._get_utc_time_of_transit(101.0, True)
            first_stars = self._get_utc_time_of_transit(96.0, False)
            three_stars = self._get_utc_time_of_transit(98.5, False)

        # shaa zmanit by gara, 1/12 of light time
        sun_hour = (sunset - sunrise) // 12
        midday = (sunset + sunrise) // 2
        mga_sunhour = (midday - first_light) / 6

        zmanim = {
            "sunrise": sunrise,
            "sunset": sunset,
            "sun_hour": sun_hour,
            "midday": midday,
            "first_light": first_light,
            "talit": talit,
            "first_stars": first_stars,
            "three_stars": three_stars,
            "plag_mincha": sunset - 1.25 * sun_hour,
            "stars_out": sunset + 18.0 * sun_hour / 60.0,
            "small_mincha": sunrise + 9.5 * sun_hour,
            "big_mincha": sunrise + 6.5 * sun_hour,
            "mga_end_shma": first_light + mga_sunhour * 3.0,
            "gra_end_shma": sunrise + sun_hour * 3.0,
            "mga_end_tfila": first_light + mga_sunhour * 4.0,
            "gra_end_tfila": sunrise + sun_hour * 4.0,
            "midnight": midday + 12 * 60.0,
        }

        return zmanim



"""Constant lookup tables for hdate modules."""

import datetime
from collections import namedtuple
from enum import Enum

READING = namedtuple("READING", "year_type, readings")

READINGS = (
    READING(
        [1725],
        (
            0,
            53,
            0,
            range(29),
            0,
            range(29, 35),
            0,
            range(35, 39),
            59,
            41,
            60,
            range(44, 51),
            61,
        ),
    ),
    READING([1703], (0, 53, 0, range(29), 0, range(29, 42), 60, range(44, 51), 61)),
    READING([1523, 523], (53, 0, range(30), 0, range(30, 51), 61)),
    READING([1501, 501], (53, 0, range(30), 0, range(30, 52))),
    READING([1317, 1227], (52, 53, range(29), 0, 0, range(29, 42), 60, range(44, 52))),
    READING(
        [1205],
        (
            52,
            53,
            range(29),
            0,
            range(29, 35),
            0,
            range(35, 39),
            59,
            41,
            60,
            range(44, 51),
            61,
        ),
    ),
    READING(
        [521, 1521],
        (53, 0, range(26), 0, 26, 56, 57, 31, 58, range(34, 42), 60, range(44, 52)),
    ),
    READING(
        [1225, 1315],
        (
            52,
            53,
            range(22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            34,
            0,
            range(35, 39),
            59,
            41,
            60,
            range(44, 51),
            61,
        ),
    ),
    READING(
        [1701],
        (
            0,
            53,
            0,
            range(22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 52),
        ),
    ),
    READING(
        [1723],
        (
            0,
            53,
            0,
            range(22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 51),
            61,
        ),
    ),
    READING(
        [1517],
        (
            53,
            0,
            range(22),
            55,
            24,
            25,
            0,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 52),
        ),
    ),
    READING(
        [703, 725],
        (0, 53, 0, 54, range(1, 29), 0, range(29, 42), 60, range(44, 51), 61),
    ),
    READING([317, 227], (52, 53, range(29), 0, range(29, 52))),
    READING([205], (52, 53, range(29), 0, range(29, 42), 60, range(44, 51), 61)),
    READING(
        [701],
        (
            0,
            53,
            0,
            54,
            range(1, 22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 52),
        ),
    ),
    READING(
        [315, 203, 225, 1203],
        (
            52,
            53,
            range(22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 51),
            61,
        ),
    ),
    READING(
        [723],
        (
            0,
            53,
            0,
            54,
            range(1, 22),
            55,
            24,
            25,
            0,
            26,
            56,
            57,
            31,
            58,
            range(34, 42),
            60,
            range(44, 51),
            61,
        ),
    ),
    READING(
        [517],
        (53, 0, range(22), 55, 24, 25, 0, 26, 56, 57, range(31, 42), 60, range(44, 52)),
    ),
)

READINGS = dict((year_type, r.readings) for r in READINGS for year_type in r.year_type)

DIGITS = (
    (" ", "א", "ב", "ג", "ד", "ה", "ו", "ז", "ח", "ט"),
    ("ט", "י", "כ", "ל", "מ", "נ", "ס", "ע", "פ", "צ"),
    (" ", "ק", "ר", "ש", "ת"),
)

LANG = namedtuple("LANG", "french, english, hebrew")
DESC = namedtuple("DESC", "long, short")

DAYS = (
    LANG(DESC("Dimanche", "Dim"), DESC("Sunday", "Sun"), DESC("ראשון", "א")),
    LANG(DESC("Lundi", "Lun"), DESC("Monday", "Mon"), DESC("שני", "ב")),
    LANG(DESC("Mardi", "Mar"), DESC("Tuesday", "Tue"), DESC("שלישי", "ג")),
    LANG(DESC("Mercredi", "Mer"), DESC("Wednesday", "Wed"), DESC("רביעי", "ד")),
    LANG(DESC("Jeudi", "Jeu"), DESC("Thursday", "Thu"), DESC("חמישי", "ה")),
    LANG(DESC("Vendredi", "Ven"), DESC("Friday", "Fri"), DESC("שישי", "ו")),
    LANG(DESC("Samedi", "Sam"), DESC("Saturday", "Sat"), DESC("שבת", "ז")),
)

PARASHAOT = (
    LANG("none", "none", "none"),
    LANG("Bereshit", "Bereshit", "בראשית"),
    LANG("Noa'h", "Noach", "נח"),
    LANG("Lekh Lekha", "Lech-Lecha", "לך לך"),
    LANG("Vayera", "Vayera", "וירא"),
    LANG("Haye Sarah", "Chayei Sara", "חיי שרה"),
    LANG("Toledot", "Toldot", "תולדות"),
    LANG("Vayetze", "Vayetzei", "ויצא"),
    LANG("Vayishla'h", "Vayishlach", "וישלח"),
    LANG("Vayeshev", "Vayeshev", "וישב"),
    LANG("Miketz", "Miketz", "מקץ"),
    LANG("Vayigash", "Vayigash", "ויגש"),
    LANG("Vaye'hi", "Vayechi", "ויחי"),
    LANG("Shemot", "Shemot", "שמות"),
    LANG("Va'era", "Vaera", "וארא"),
    LANG("Bo", "Bo", "בא"),
    LANG("Beshalakh", "Beshalach", "בשלח"),
    LANG("Yitro", "Yitro", "יתרו"),
    LANG("Mishpatim", "Mishpatim", "משפטים"),
    LANG("Teroumah", "Terumah", "תרומה"),
    LANG("Tetzave", "Tetzaveh", "תצוה"),
    LANG("Ki Tissa", "Ki Tisa", "כי תשא"),
    LANG("Vayaqhel", "Vayakhel", "ויקהל"),
    LANG("Peqoudei", "Pekudei", "פקודי"),
    LANG("Vayikra", "Vayikra", "ויקרא"),
    LANG("Tzav", "Tzav", "צו"),
    LANG("Shemini", "Shmini", "שמיני"),
    LANG("Tazria", "Tazria", "תזריע"),
    LANG("Metzora", "Metzora", "מצורע"),
    LANG("A'harei Mot", "Achrei Mot", "אחרי מות"),
    LANG("Kedoshim", "Kedoshim", "קדושים"),
    LANG("Emor", "Emor", "אמור"),
    LANG("Behar", "Behar", "בהר"),
    LANG("Be'houkotai", "Bechukotai", "בחוקתי"),
    LANG("Bamidbar", "Bamidbar", "במדבר"),
    LANG("Nasso", "Nasso", "נשא"),
    LANG("Beha'alotkha", "Beha'alotcha", "בהעלתך"),
    LANG("Shla'h lekha", "Sh'lach", "שלח"),
    LANG("Kora'h", "Korach", "קרח"),
    LANG("Houkat", "Chukat", "חקת"),
    LANG("Balak", "Balak", "בלק"),
    LANG("Pin'has", "Pinchas", "פנחס"),
    LANG("Matot", "Matot", "מטות"),
    LANG("Massei", "Masei", "מסעי"),
    LANG("Devarim", "Devarim", "דברים"),
    LANG("Va'et'hanan", "Vaetchanan", "ואתחנן"),
    LANG("Eikev", "Eikev", "עקב"),
    LANG("Re'eh", "Re'eh", "ראה"),
    LANG("Shoftim", "Shoftim", "שופטים"),
    LANG("Ki Tetze", "Ki Teitzei", "כי תצא"),
    LANG("Ki Tavo", "Ki Tavo", "כי תבוא"),
    LANG("Nitzavim", "Nitzavim", "נצבים"),
    LANG("Vayelekh", "Vayeilech", "וילך"),
    LANG("Haazinou", "Ha'Azinu", "האזינו"),
    LANG("Vezot Haberakha", "Vezot Habracha", "וזאת הברכה"),
    LANG("Vayaqhel-Peqoudei", "Vayakhel-Pekudei", "ויקהל-פקודי"),
    LANG("Tazria-Metzora", "Tazria-Metzora", "תזריע-מצורע"),
    LANG("A'harei Mot-Kedoshim", "Achrei Mot-Kedoshim", "אחרי מות-קדושים"),
    LANG("Behar-Be'houkotai", "Behar-Bechukotai", "בהר-בחוקתי"),
    LANG("Houkat-Balak", "Chukat-Balak", "חוקת-בלק"),
    LANG("Matot-Massei", "Matot-Masei", "מטות מסעי"),
    LANG("Nitzavim-Vayelekh", "Nitzavim-Vayeilech", "נצבים-וילך"),
)

MONTHS = (
    LANG("Tishri", "Tishrei", "תשרי"),
    LANG("Heshvan", "Marcheshvan", "מרחשוון"),
    LANG("Kislev", "Kislev", "כסלו"),
    LANG("Tevet", "Tevet", "טבת"),
    LANG("Shvat", "Sh'vat", "שבט"),
    LANG("Adar", "Adar", "אדר"),
    LANG("Nissan", "Nisan", "ניסן"),
    LANG("Iyar", "Iyyar", "אייר"),
    LANG("Sivan", "Sivan", "סיון"),
    LANG("Tamouz", "Tammuz", "תמוז"),
    LANG("Av", "Av", "אב"),
    LANG("Eloul", "Elul", "אלול"),
    LANG("Adar I", "Adar I", "אדר א"),
    LANG("Adar II", "Adar II", "אדר ב"),
)


class Months(Enum):
    """Enum class for the Hebrew months."""

    TISHREI = 1
    MARCHESHVAN = 2
    KISLEV = 3
    TEVET = 4
    SHVAT = 5
    ADAR = 6
    NISAN = 7
    IYYAR = 8
    SIVAN = 9
    TAMMUZ = 10
    AV = 11
    ELUL = 12
    ADAR_I = 13
    ADAR_II = 14


def year_is_after(year):
    """
    Return a lambda function.

    Lambda checks that a given HDate object's hebrew year is after the
    requested year.
    """
    return lambda x: x.hdate.year > year


def year_is_before(year):
    """
    Return a lambda function.

    Lambda checks that a given HDate object's hebrew year is before the
    requested year.
    """
    return lambda x: x.hdate.year < year


def move_if_not_on_dow(original, replacement, dow_not_orig, dow_replacement):
    """
    Return a lambda function.

    Lambda checks that either the original day does not fall on a given
    weekday, or that the replacement day does fall on the expected weekday.
    """
    return lambda x: (
        (x.hdate.day == original and x.gdate.weekday() != dow_not_orig)
        or (x.hdate.day == replacement and x.gdate.weekday() == dow_replacement)
    )


def correct_adar():
    """
    Return a lambda function.

    Lambda checks that the value of the month returned is correct depending on whether
    it's a leap year.
    """
    return lambda x: (
        (x.hdate.month == Months.ADAR and not x.is_leap_year)
        or (x.hdate.month in [Months.ADAR_I, Months.ADAR_II] and x.is_leap_year)
    )


HOLIDAY = namedtuple(
    "HOLIDAY",
    ["type", "name", "date", "israel_diaspora", "date_functions_list", "description"],
)


class HolidayTypes(Enum):
    """Container class for holiday type integer mappings."""

    UNKNOWN = 0
    YOM_TOV = 1
    EREV_YOM_TOV = 2
    HOL_HAMOED = 3
    MELACHA_PERMITTED_HOLIDAY = 4
    FAST_DAY = 5
    MODERN_HOLIDAY = 6
    MINOR_HOLIDAY = 7
    MEMORIAL_DAY = 8
    ISRAEL_NATIONAL_HOLIDAY = 9


HOLIDAYS = (
    HOLIDAY(HolidayTypes.UNKNOWN, "", (), "", [], LANG("", "", DESC("", ""))),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "erev_rosh_hashana",
        (29, Months.ELUL),
        "",
        [],
        LANG(
            "Veille de Rosh Hashana",
            "Erev Rosh Hashana",
            DESC("ערב ראש השנה", 'ערב ר"ה'),
        ),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "rosh_hashana_i",
        (1, Months.TISHREI),
        "",
        [],
        LANG("Rosh Hashana I", "Rosh Hashana I", DESC("א' ראש השנה", 'א ר"ה')),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "rosh_hashana_ii",
        (2, Months.TISHREI),
        "",
        [],
        LANG("Rosh Hashana II", "Rosh Hashana II", DESC("ב' ראש השנה", "ב' ר\"ה")),
    ),
    HOLIDAY(
        HolidayTypes.FAST_DAY,
        "tzom_gedaliah",
        ([3, 4], Months.TISHREI),
        "",
        [move_if_not_on_dow(3, 4, 5, 6)],
        LANG("Jeûne de Guedalia", "Tzom Gedaliah", DESC("צום גדליה", "צום גדליה")),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "erev_yom_kippur",
        (9, Months.TISHREI),
        "",
        [],
        LANG("Veille de Yom Kipour", "Erev Yom Kippur", DESC('עיוה"כ', 'עיוה"כ')),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "yom_kippur",
        (10, Months.TISHREI),
        "",
        [],
        LANG("Yom Kipour", "Yom Kippur", DESC("יום הכפורים", 'יוה"כ')),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "erev_sukkot",
        (14, Months.TISHREI),
        "",
        [],
        LANG("Veille de Souccot", "Erev Sukkot", DESC("ערב סוכות", "ערב סוכות")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "sukkot",
        (15, Months.TISHREI),
        "",
        [],
        LANG("Souccot", "Sukkot", DESC("סוכות", "סוכות")),
    ),
    HOLIDAY(
        HolidayTypes.HOL_HAMOED,
        "hol_hamoed_sukkot",
        (16, Months.TISHREI),
        "ISRAEL",
        "",
        LANG(
            "Hol hamoed Souccot",
            "Hol hamoed Sukkot",
            DESC("חול המועד סוכות", 'חוה"מ סוכות'),
        ),
    ),
    HOLIDAY(
        HolidayTypes.HOL_HAMOED,
        "hol_hamoed_sukkot",
        ([17, 18, 19, 20], Months.TISHREI),
        "",
        "",
        LANG(
            "Hol hamoed Souccot",
            "Hol hamoed Sukkot",
            DESC("חול המועד סוכות", 'חוה"מ סוכות'),
        ),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "hoshana_raba",
        (21, Months.TISHREI),
        "",
        [],
        LANG("Hoshaâna Rabba", "Hoshana Raba", DESC("הושענא רבה", 'הוש"ר')),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "simchat_torah",
        (23, Months.TISHREI),
        "DIASPORA",
        [],
        LANG("Simhat Torah", "Simchat Torah", DESC("שמחת תורה", 'שמח"ת')),
    ),
    HOLIDAY(
        HolidayTypes.MELACHA_PERMITTED_HOLIDAY,
        "chanukah",
        (list(range(25, 30)), Months.KISLEV),
        "",
        [],
        LANG("Hanouka", "Chanukah", DESC("חנוכה", "חנוכה")),
    ),
    HOLIDAY(
        HolidayTypes.MELACHA_PERMITTED_HOLIDAY,
        "chanukah",
        ([1, 2, 3], Months.TEVET),
        "",
        [
            lambda x: (
                (x.short_kislev() and x.hdate.day == 3) or (x.hdate.day in [1, 2])
            )
        ],
        LANG("Hanouka", "Chanukah", DESC("חנוכה", "חנוכה")),
    ),
    HOLIDAY(
        HolidayTypes.FAST_DAY,
        "asara_btevet",
        (10, Months.TEVET),
        "",
        [],
        LANG("10 Tevet", "Asara B'Tevet", DESC("צום עשרה בטבת", "י' בטבת")),
    ),
    HOLIDAY(
        HolidayTypes.MINOR_HOLIDAY,
        "tu_bshvat",
        (15, Months.SHVAT),
        "",
        [],
        LANG("Tou Bi Shvat", "Tu B'Shvat", DESC('ט"ו בשבט', 'ט"ו בשבט')),
    ),
    HOLIDAY(
        HolidayTypes.FAST_DAY,
        "taanit_esther",
        ([11, 13], [Months.ADAR, Months.ADAR_II]),
        "",
        [move_if_not_on_dow(13, 11, 5, 3), correct_adar()],
        LANG("Jeûne d'Esther", "Ta'anit Esther", DESC("תענית אסתר", "תענית אסתר")),
    ),
    HOLIDAY(
        HolidayTypes.MELACHA_PERMITTED_HOLIDAY,
        "purim",
        (14, [Months.ADAR, Months.ADAR_II]),
        "",
        [correct_adar()],
        LANG("Pourim", "Purim", DESC("פורים", "פורים")),
    ),
    HOLIDAY(
        HolidayTypes.MELACHA_PERMITTED_HOLIDAY,
        "shushan_purim",
        (15, [Months.ADAR, Months.ADAR_II]),
        "",
        [correct_adar()],
        LANG("Pourim Shoushan", "Shushan Purim", DESC("שושן פורים", "שושן פורים")),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "erev_pesach",
        (14, Months.NISAN),
        "",
        [],
        LANG("Veille de Pessah", "Erev Pesach", DESC("ערב פסח", "ערב פסח")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "pesach",
        (15, Months.NISAN),
        "",
        "",
        LANG("Pessah", "Pesach", DESC("פסח", "פסח")),
    ),
    HOLIDAY(
        HolidayTypes.HOL_HAMOED,
        "hol_hamoed_pesach",
        (16, Months.NISAN),
        "ISRAEL",
        [],
        LANG(
            "Hol hamoed Pessah", "Hol hamoed Pesach", DESC("חול המועד פסח", 'חוה"מ פסח')
        ),
    ),
    HOLIDAY(
        HolidayTypes.HOL_HAMOED,
        "hol_hamoed_pesach",
        ([17, 18, 19], Months.NISAN),
        "",
        [],
        LANG(
            "Hol hamoed Pessah", "Hol hamoed Pesach", DESC("חול המועד פסח", 'חוה"מ פסח')
        ),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "hol_hamoed_pesach",
        (20, Months.NISAN),
        "",
        [],
        LANG(
            "Hol hamoed Pessah", "Hol hamoed Pesach", DESC("חול המועד פסח", 'חוה"מ פסח')
        ),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "pesach_vii",
        (21, Months.NISAN),
        "",
        [],
        LANG("Pessah VII", "Pesach VII", DESC("שביעי פסח", "ז' פסח")),
    ),
    HOLIDAY(
        HolidayTypes.MODERN_HOLIDAY,
        "yom_haatzmaut",
        ([3, 4, 5], Months.IYYAR),
        "",
        [
            year_is_after(5708),
            year_is_before(5764),
            move_if_not_on_dow(5, 4, 4, 3) or move_if_not_on_dow(5, 3, 5, 3),
        ],
        LANG("Yom HaAtsmaout", "Yom HaAtzma'ut", DESC("יום העצמאות", "יום העצמאות")),
    ),
    HOLIDAY(
        HolidayTypes.MODERN_HOLIDAY,
        "yom_haatzmaut",
        ([3, 4, 5, 6], Months.IYYAR),
        "",
        [
            year_is_after(5763),
            move_if_not_on_dow(5, 4, 4, 3)
            or move_if_not_on_dow(5, 3, 5, 3)
            or move_if_not_on_dow(5, 6, 0, 1),
        ],
        LANG("Yom HaAtsmaout", "Yom HaAtzma'ut", DESC("יום העצמאות", "יום העצמאות")),
    ),
    HOLIDAY(
        HolidayTypes.MINOR_HOLIDAY,
        "lag_bomer",
        (18, Months.IYYAR),
        "",
        [],
        LANG("Lag Ba Omer", "Lag B'Omer", DESC('ל"ג בעומר', 'ל"ג בעומר')),
    ),
    HOLIDAY(
        HolidayTypes.EREV_YOM_TOV,
        "erev_shavuot",
        (5, Months.SIVAN),
        "",
        [],
        LANG("Veille de Shavouot", "Erev Shavuot", DESC("ערב שבועות", "ערב שבועות")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "shavuot",
        (6, Months.SIVAN),
        "",
        [],
        LANG("Shavouot", "Shavuot", DESC("שבועות", "שבועות")),
    ),
    HOLIDAY(
        HolidayTypes.FAST_DAY,
        "tzom_tammuz",
        ([17, 18], Months.TAMMUZ),
        "",
        [move_if_not_on_dow(17, 18, 5, 6)],
        LANG(
            "Jeûne du 17 Tamouz", "Tzom Tammuz", DESC("צום שבעה עשר בתמוז", "צום תמוז")
        ),
    ),
    HOLIDAY(
        HolidayTypes.FAST_DAY,
        "tisha_bav",
        ([9, 10], Months.AV),
        "",
        [move_if_not_on_dow(9, 10, 5, 6)],
        LANG("Tisha Be'Av", "Tish'a B'Av", DESC("תשעה באב", "ט' באב")),
    ),
    HOLIDAY(
        HolidayTypes.MINOR_HOLIDAY,
        "tu_bav",
        (15, Months.AV),
        "",
        [],
        LANG("Tou be'Av", "Tu B'Av", DESC('ט"ו באב', 'ט"ו באב')),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "yom_hashoah",
        ([26, 27, 28], Months.NISAN),
        "",
        [
            move_if_not_on_dow(27, 28, 6, 0) or move_if_not_on_dow(27, 26, 4, 3),
            year_is_after(5718),
        ],
        LANG("Yom HaShoah", "Yom HaShoah", DESC("יום השואה", "יום השואה")),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "yom_hazikaron",
        ([2, 3, 4], Months.IYYAR),
        "",
        [
            year_is_after(5708),
            year_is_before(5764),
            move_if_not_on_dow(4, 3, 3, 2) or move_if_not_on_dow(4, 2, 4, 2),
        ],
        LANG("Yom haZicaron", "Yom HaZikaron", DESC("יום הזכרון", "יום הזכרון")),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "yom_hazikaron",
        ([2, 3, 4, 5], Months.IYYAR),
        "",
        [
            year_is_after(5763),
            move_if_not_on_dow(4, 3, 3, 2)
            or move_if_not_on_dow(4, 2, 4, 2)
            or move_if_not_on_dow(4, 5, 6, 0),
        ],
        LANG("Yom haZicaron", "Yom HaZikaron", DESC("יום הזכרון", "יום הזכרון")),
    ),
    HOLIDAY(
        HolidayTypes.MODERN_HOLIDAY,
        "yom_yerushalayim",
        (28, Months.IYYAR),
        "",
        [year_is_after(5727)],
        LANG("Yom Yeroushalaïm", "Yom Yerushalayim", DESC("יום ירושלים", "יום י-ם")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "shmini_atzeret",
        (22, Months.TISHREI),
        "",
        [],
        LANG("Shemini Atseret", "Shmini Atzeret", DESC("שמיני עצרת", "שמיני עצרת")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "pesach_viii",
        (22, Months.NISAN),
        "DIASPORA",
        [],
        LANG("Pessah VIII", "Pesach VIII", DESC("אחרון של פסח", "אחרון של פסח")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "shavuot_ii",
        (7, Months.SIVAN),
        "DIASPORA",
        [],
        LANG("Shavouot II", "Shavuot II", DESC("שני של שבועות", "ב' שבועות")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "sukkot_ii",
        (16, Months.TISHREI),
        "DIASPORA",
        [],
        LANG("Souccot II", "Sukkot II", DESC("שני של סוכות", "ב' סוכות")),
    ),
    HOLIDAY(
        HolidayTypes.YOM_TOV,
        "pesach_ii",
        (16, Months.NISAN),
        "DIASPORA",
        [],
        LANG("Pessah II", "Pesach II", DESC("שני של פסח", "ב' פסח")),
    ),
    HOLIDAY(
        HolidayTypes.ISRAEL_NATIONAL_HOLIDAY,
        "family_day",
        (30, Months.SHVAT),
        "ISRAEL",
        [year_is_after(5734)],
        LANG("Family Day", "Family Day", DESC("יום המשפחה", "יום המשפחה")),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "memorial_day_unknown",
        (7, [Months.ADAR, Months.ADAR_II]),
        "ISRAEL",
        [correct_adar()],
        LANG(
            "Jour du souvenir",
            "Memorial day for fallen whose place of burial is unknown",
            DESC("יום זכרון...", "יום זכרון..."),
        ),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "rabin_memorial_day",
        ([11, 12], Months.MARCHESHVAN),
        "ISRAEL",
        [move_if_not_on_dow(12, 11, 4, 3), year_is_after(5757)],
        LANG(
            "Yitzhak Rabin memorial day",
            "Yitzhak Rabin memorial day",
            DESC("יום הזכרון ליצחק רבין", "יום הזכרון ליצחק רבין"),
        ),
    ),
    HOLIDAY(
        HolidayTypes.MEMORIAL_DAY,
        "zeev_zhabotinsky_day",
        (29, Months.TAMMUZ),
        "ISRAEL",
        [year_is_after(5764)],
        LANG(
            "Zeev Zhabotinsky day",
            "Zeev Zhabotinsky day",
            DESC("יום ז'בוטינסקי", "יום ז'בוטינסקי"),
        ),
    ),
)

ZMAN = namedtuple("ZMAN", "zman, description")
ZMANIM = (
    ZMAN("first_light", LANG("Alot HaShahar", "Alot HaShachar", "עלות השחר")),
    ZMAN(
        "talit",
        LANG("Début Talit & Tefilin ", "Talit & Tefilin's time", "זמן טלית ותפילין"),
    ),
    ZMAN("sunrise", LANG("Lever du jour", "Sunrise", "הנץ החמה")),
    ZMAN("mga_end_shma", LANG('Shema MG"A', 'Shema EOT MG"A', 'סוף זמן ק"ש מג"א')),
    ZMAN("gra_end_shma", LANG('Shema GR"A', 'Shema EOT GR"A', 'סוף זמן ק"ש גר"א')),
    ZMAN("mga_end_tfila", LANG('Tefila MG"A', 'Tefila EOT MG"A', 'סוף זמן תפילה מג"א')),
    ZMAN("gra_end_tfila", LANG('Tefila GR"A', 'Tefila EOT GR"A', 'סוף זמן תפילה גר"א')),
    ZMAN("midday", LANG("Hatsot", "Midday", "חצות היום")),
    ZMAN("big_mincha", LANG("Minha Guedola", "Big Mincha", "מנחה גדולה")),
    ZMAN("small_mincha", LANG("Minha Qetana", "Small Mincha", "מנחה קטנה")),
    ZMAN("plag_mincha", LANG("Plag haMinha", "Plag Mincha", "פלג המנחה")),
    ZMAN("sunset", LANG("Shqiat", "Sunset", "שקיעה")),
    ZMAN("first_stars", LANG("Tzeit haCokhavim", "First stars", "צאת הכוכבים")),
    ZMAN("midnight", LANG("Hatsot laïla", "Midnight", "חצות הלילה")),
)

# The first few cycles were only 2702 blatt. After that it became 2711. Even with
# that, the math doesn't play nicely with the dates before the 11th cycle :(
# From cycle 11 onwards, it was simple and sequential
DAF_YOMI_CYCLE_11_START = datetime.date(1997, 9, 29)
MESECHTA = namedtuple("MESECHTA", ["name", "pages"])
DAF_YOMI_MESECHTOS = (
    MESECHTA(LANG("Berachos", "Berachos", "ברכות"), 63),
    MESECHTA(LANG("Shabbos", "Shabbos", "שבת"), 156),
    MESECHTA(LANG("Eruvin", "Eruvin", "עירובין"), 104),
    MESECHTA(LANG("Pesachim", "Pesachim", "פסחים"), 120),
    MESECHTA(LANG("Shekalim", "Shekalim", "שקלים"), 21),
    MESECHTA(LANG("Yoma", "Yoma", "יומא"), 87),
    MESECHTA(LANG("Succah", "Succah", "סוכה"), 55),
    MESECHTA(LANG("Beitzah", "Beitzah", "ביצה"), 39),
    MESECHTA(LANG("Rosh Hashanah", "Rosh Hashanah", "ראש השנה"), 34),
    MESECHTA(LANG("Taanis", "Taanis", "תענית"), 30),
    MESECHTA(LANG("Megillah", "Megillah", "מגילה"), 31),
    MESECHTA(LANG("Moed Katan", "Moed Katan", "מועד קטן"), 28),
    MESECHTA(LANG("Chagigah", "Chagigah", "חגיגה"), 26),
    MESECHTA(LANG("Yevamos", "Yevamos", "יבמות"), 121),
    MESECHTA(LANG("Kesubos", "Kesubos", "כתובות"), 111),
    MESECHTA(LANG("Nedarim", "Nedarim", "נדרים"), 90),
    MESECHTA(LANG("Nazir", "Nazir", "נזיר"), 65),
    MESECHTA(LANG("Sotah", "Sotah", "סוטה"), 48),
    MESECHTA(LANG("Gittin", "Gittin", "גיטין"), 89),
    MESECHTA(LANG("Kiddushin", "Kiddushin", "קידושין"), 81),
    MESECHTA(LANG("Bava Kamma", "Bava Kamma", "בבא קמא"), 118),
    MESECHTA(LANG("Bava Metzia", "Bava Metzia", "בבא מציעא"), 118),
    MESECHTA(LANG("Bava Basra", "Bava Basra", "בבא בתרא"), 175),
    MESECHTA(LANG("Sanhedrin", "Sanhedrin", "סנהדרין"), 112),
    MESECHTA(LANG("Makkos", "Makkos", "מכות"), 23),
    MESECHTA(LANG("Shevuos", "Shevuos", "שבועות"), 48),
    MESECHTA(LANG("Avodah Zarah", "Avodah Zarah", "עבודה זרה"), 75),
    MESECHTA(LANG("Horayos", "Horayos", "הוריות"), 13),
    MESECHTA(LANG("Zevachim", "Zevachim", "זבחים"), 119),
    MESECHTA(LANG("Menachos", "Menachos", "מנחות"), 109),
    MESECHTA(LANG("Chullin", "Chullin", "חולין"), 141),
    MESECHTA(LANG("Bechoros", "Bechoros", "בכורות"), 60),
    MESECHTA(LANG("Arachin", "Arachin", "ערכין"), 33),
    MESECHTA(LANG("Temurah", "Temurah", "תמורה"), 33),
    MESECHTA(LANG("Kereisos", "Kereisos", "כריתות"), 27),
    MESECHTA(LANG("Meilah", "Meilah", "מעילה"), 36),
    MESECHTA(LANG("Niddah", "Niddah", "נדה"), 72),
)

DAF_YOMI_TOTAL_PAGES = sum(mesechta.pages for mesechta in DAF_YOMI_MESECHTOS)



"""
Jewish calendrical date and times for a given location.

HDate calculates and generates a representation either in English or Hebrew
of the Jewish calendrical date and times for a given location
"""

from hdate.date import HDate
from hdate.hebrew_date import HebrewDate
from hdate.htables import HolidayTypes
from hdate.location import Location
from hdate.zmanim import Zmanim

__all__ = ["HDate", "Zmanim", "HebrewDate", "Location", "HolidayTypes"]



"""Fixtures for py.test."""

import datetime
import random
from calendar import isleap

import pytest

import hdate


@pytest.fixture
def random_date():
    """Generate a random valid date."""
    year = random.randint(400, 2500)
    month = random.randint(1, 12)
    maxday = 31 if month in [1, 3, 5, 7, 8, 10, 12] else 30
    if month == 2:
        if isleap(year):
            maxday = 29
        else:
            maxday = 28
    day = random.randint(1, maxday)
    return datetime.date(year, month, day)


@pytest.fixture
def rand_hdate(random_date):  # pylint: disable=redefined-outer-name
    """Given a random date, generate a random HDate."""
    return hdate.HDate(random_date)



"""Test the conversion functions."""

import pytest

from hdate import converters as conv
from hdate.hebrew_date import HebrewDate
from hdate.htables import Months


class TestConverters:
    """Tests for converting one date type to another."""

    def test_gdate_to_gdate(self, random_date):
        """ "Transform Gregorian date to Gregorian date."""
        assert conv.jdn_to_gdate(conv.gdate_to_jdn(random_date)) == random_date

    YEARS_PSHUTA = [5753, 5762, 5756]
    YEARS_MEUBERET = [5749, 5755, 5760]

    SIMPLE_MONTHS = {
        29: [Months.TEVET, Months.IYYAR, Months.TAMMUZ, Months.ELUL],
        30: [Months.TISHREI, Months.SHVAT, Months.NISAN, Months.SIVAN, Months.AV],
    }

    @pytest.mark.parametrize("year", YEARS_PSHUTA)
    def test_hdate_to_hdate_pshuta_simple(self, year):
        """Transform Hebrew date to Hebrew date (single Adar)."""
        for days_in_month, months in self.SIMPLE_MONTHS.items():
            for month in months:
                for day in range(1, days_in_month + 1):
                    date = HebrewDate(year, month, day)
                    assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_PSHUTA)
    def test_hdate_to_hdate_pshuta_adar(self, year):
        """Transform Hebrew date to Hebrew date (in Adar)."""
        for day in range(1, 29 + 1):
            date = HebrewDate(year, Months.ADAR, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_PSHUTA)
    def test_hdate_to_hdate_pshuta_heshvan(self, year):
        """Transform Hebrew date to hebrew date (Heshvan)."""
        days_in_month = 29 if not year == 5756 else 30
        for day in range(1, days_in_month + 1):
            date = HebrewDate(year, Months.MARCHESHVAN, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_PSHUTA)
    def test_hdate_to_hdate_pshuta_kislev(self, year):
        """Transform Hebrew date to hebrew date (Kislev)."""
        days_in_month = 30 if not year == 5753 else 29
        for day in range(1, days_in_month + 1):
            date = HebrewDate(year, Months.KISLEV, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_MEUBERET)
    def test_hdate_to_hdate_meuberet_simple(self, year):
        """Transform Hebrew date to hebrew date (two Adars)."""
        for days_in_month, months in self.SIMPLE_MONTHS.items():
            for month in months:
                for day in range(1, days_in_month + 1):
                    date = HebrewDate(year, month, day)
                    assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_MEUBERET)
    def test_hdate_to_hdate_meuberet_adar(self, year):
        """Transform Hebrew date to hebrew date (two Adars - test Adar)."""
        for day in range(1, 30 + 1):
            date = HebrewDate(year, Months.ADAR_I, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date
        for day in range(1, 29 + 1):
            date = HebrewDate(year, Months.ADAR_II, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_MEUBERET)
    def test_hdate_to_hdate_meuberet_heshvan(self, year):
        """Transform Hebrew date to hebrew date (two Adars - test Heshvan)."""
        days_in_month = 29 if not year == 5760 else 30
        for day in range(1, days_in_month + 1):
            date = HebrewDate(year, Months.MARCHESHVAN, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date

    @pytest.mark.parametrize("year", YEARS_MEUBERET)
    def test_hdate_to_hdate_meuberet_kislev(self, year):
        """Transform Hebrew date to hebrew date (two Adars - test Kislev)."""
        days_in_month = 30 if not year == 5749 else 29
        for day in range(1, days_in_month + 1):
            date = HebrewDate(year, Months.KISLEV, day)
            assert conv.jdn_to_hdate(conv.hdate_to_jdn(date)) == date



"""
Jewish calendrical date and times for a given location.

HDate calculates and generates a representation either in English or Hebrew
of the Jewish calendrical date and times for a given location
"""

import datetime
import logging
from itertools import chain, product

from hdate import converters as conv
from hdate import htables
from hdate.hebrew_date import HebrewDate
from hdate.htables import HolidayTypes, Months

_LOGGER = logging.getLogger(__name__)
# pylint: disable=too-many-public-methods


class HDate:
    """
    Hebrew date class.

    Supports converting from Gregorian and Julian to Hebrew date.
    """

    def __init__(
        self, gdate=datetime.date.today(), diaspora=False, hebrew=True, heb_date=None
    ):
        """Initialize the HDate object."""
        # Create private variables
        self._hdate = None
        self._gdate = None
        self._last_updated = None

        # Assign values
        # Keep hdate after gdate assignment so as not to cause recursion error
        if heb_date is None:
            self.gdate = gdate
            self.hdate = None
        else:
            self.gdate = None
            self.hdate = heb_date
        self.hebrew = hebrew
        self.diaspora = diaspora

    def __str__(self):
        """Return a full Unicode representation of HDate."""
        result = (
            f"{'יום ' if self.hebrew else ''}"
            f"{htables.DAYS[self.dow - 1][self.hebrew + 1][0]} "
            f"{hebrew_number(self.hdate.day, hebrew=self.hebrew)} "
            f"{'ב' if self.hebrew else ''}"
            f"{htables.MONTHS[self.hdate.month.value - 1][self.hebrew + 1]} "
            f"{hebrew_number(self.hdate.year, hebrew=self.hebrew)}"
        )

        if 0 < self.omer_day < 50:
            result = f"{result} {hebrew_number(self.omer_day, hebrew=self.hebrew)}"
            result = f"{result} {'בעומר' if self.hebrew else ' in the Omer'}"

        if self.holiday_description:
            result = f"{result} {self.holiday_description}"
        return result

    def __repr__(self):
        """Return a representation of HDate for programmatic use."""
        return (
            f"HDate(gdate={self.gdate!r}, diaspora={self.diaspora}, "
            f"hebrew={self.hebrew})"
        )

    def __lt__(self, other):
        """Implement the less-than operator."""
        assert isinstance(other, HDate)
        return self.gdate < other.gdate

    def __le__(self, other):
        """Implement the less-than or equal operator."""
        return not other < self

    def __gt__(self, other):
        """Implement the greater-than operator."""
        return other < self

    def __ge__(self, other):
        """Implement the greater than or equal operator."""
        return not self < other

    @property
    def hdate(self):
        """Return the hebrew date."""
        if self._last_updated == "hdate":
            return self._hdate
        return conv.jdn_to_hdate(self._jdn)

    @hdate.setter
    def hdate(self, date):
        """Set the dates of the HDate object based on a given Hebrew date."""
        # Sanity checks
        if date is None and isinstance(self.gdate, datetime.date):
            # Calculate the value since gdate has been set
            date = self.hdate

        if not isinstance(date, HebrewDate):
            raise TypeError(f"date: {date} is not of type HebrewDate")
        if not 0 < date.day < 31:
            raise ValueError(f"day ({date.day}) legal values are 1-31")

        self._last_updated = "hdate"
        self._hdate = date

    @property
    def gdate(self):
        """Return the Gregorian date for the given Hebrew date object."""
        if self._last_updated == "gdate":
            return self._gdate
        return conv.jdn_to_gdate(self._jdn)

    @gdate.setter
    def gdate(self, date):
        """Set the Gregorian date for the given Hebrew date object."""
        self._last_updated = "gdate"
        self._gdate = date

    @property
    def _jdn(self):
        """Return the Julian date number for the given date."""
        if self._last_updated == "gdate":
            return conv.gdate_to_jdn(self.gdate)
        return conv.hdate_to_jdn(self.hdate)

    @property
    def hebrew_date(self):
        """Return the hebrew date string."""
        return (
            f"{hebrew_number(self.hdate.day, hebrew=self.hebrew)} "  # Day
            f"{htables.MONTHS[self.hdate.month.value - 1][self.hebrew + 1]} "  # Month
            f"{hebrew_number(self.hdate.year, hebrew=self.hebrew)}"  # Year
        )

    @property
    def parasha(self):
        """Return the upcoming parasha."""
        return htables.PARASHAOT[self.get_reading()][self.hebrew + 1]

    @property
    def holiday_description(self):
        """
        Return the holiday description.

        In case none exists will return None.
        """
        entry = self._holiday_entry()
        desc = entry.description
        return desc.hebrew.long if self.hebrew else desc.english

    @property
    def is_shabbat(self):
        """Return True if this date is Shabbat, specifically Saturday.

        Returns False on Friday because the HDate object has no notion of time.
        For more detailed nuance, use the Zmanim object.
        """
        return self.gdate.weekday() == 5

    @property
    def is_holiday(self):
        """Return True if this date is a holiday (any kind)."""
        return self.holiday_type != HolidayTypes.UNKNOWN

    @property
    def is_yom_tov(self):
        """Return True if this date is a Yom Tov."""
        return self.holiday_type == HolidayTypes.YOM_TOV

    @property
    def is_leap_year(self):
        """Return True if this date's year is a leap year."""
        return self.hdate.year % 19 in [0, 3, 6, 8, 11, 14, 17]

    @property
    def holiday_type(self):
        """Return the holiday type if exists."""
        entry = self._holiday_entry()
        return entry.type

    @property
    def holiday_name(self):
        """Return the holiday name which is good for programmatic use."""
        entry = self._holiday_entry()
        return entry.name

    def _holiday_entry(self):
        """Return the abstract holiday information from holidays table."""
        holidays_list = self.get_holidays_for_year()
        holidays_list = [
            holiday
            for holiday, holiday_hdate in holidays_list
            if holiday_hdate.hdate == self.hdate
        ]
        assert len(holidays_list) <= 1

        # If anything is left return it, otherwise return the "NULL" holiday
        return holidays_list[0] if holidays_list else htables.HOLIDAYS[0]

    def short_kislev(self):
        """Return whether this year has a short Kislev or not."""
        return self.year_size() in [353, 383]

    @property
    def dow(self):
        """Return Hebrew day of week Sunday = 1, Saturday = 7."""
        # datetime weekday maps Monday->0, Sunday->6; this remaps to Sunday->1.
        return self.gdate.weekday() + 2 if self.gdate.weekday() != 6 else 1

    def year_size(self):
        """Return the size of the given Hebrew year."""
        return conv.get_size_of_hebrew_year(self.hdate.year)

    def rosh_hashana_dow(self):
        """Return the Hebrew day of week for Rosh Hashana."""
        jdn = conv.hdate_to_jdn(HebrewDate(self.hdate.year, Months.TISHREI, 1))
        return (jdn + 1) % 7 + 1

    def pesach_dow(self):
        """Return the first day of week for Pesach."""
        jdn = conv.hdate_to_jdn(HebrewDate(self.hdate.year, Months.NISAN, 15))
        return (jdn + 1) % 7 + 1

    @property
    def omer_day(self):
        """Return the day of the Omer."""
        first_omer_day = HebrewDate(self.hdate.year, Months.NISAN, 16)
        omer_day = self._jdn - conv.hdate_to_jdn(first_omer_day) + 1
        if not 0 < omer_day < 50:
            return 0
        return omer_day

    @property
    def daf_yomi_repr(self):
        """Return a tuple of mesechta and daf."""
        days_since_start_cycle_11 = (self.gdate - htables.DAF_YOMI_CYCLE_11_START).days
        page_number = days_since_start_cycle_11 % (htables.DAF_YOMI_TOTAL_PAGES)
        for mesechta in htables.DAF_YOMI_MESECHTOS:
            if page_number >= mesechta.pages:
                page_number -= mesechta.pages
            else:
                break
        daf_number = page_number + 2
        return mesechta, daf_number

    @property
    def daf_yomi(self):
        """Return a string representation of the daf yomi."""
        mesechta, daf_number = self.daf_yomi_repr
        if self.hebrew:
            mesechta_name = mesechta.name.hebrew
        else:
            mesechta_name = mesechta.name.english
        daf = hebrew_number(daf_number, self.hebrew, short=True)
        return f"{mesechta_name} {daf}"

    @property
    def next_day(self):
        """Return the HDate for the next day."""
        return HDate(self.gdate + datetime.timedelta(1), self.diaspora, self.hebrew)

    @property
    def previous_day(self):
        """Return the HDate for the previous day."""
        return HDate(self.gdate + datetime.timedelta(-1), self.diaspora, self.hebrew)

    @property
    def upcoming_shabbat(self):
        """Return the HDate for either the upcoming or current Shabbat.

        If it is currently Shabbat, returns the HDate of the Saturday.
        """
        if self.is_shabbat:
            return self
        # If it's Sunday, fast forward to the next Shabbat.
        saturday = self.gdate + datetime.timedelta((12 - self.gdate.weekday()) % 7)
        return HDate(saturday, diaspora=self.diaspora, hebrew=self.hebrew)

    @property
    def upcoming_shabbat_or_yom_tov(self):
        """Return the HDate for the upcoming or current Shabbat or Yom Tov.

        If it is currently Shabbat, returns the HDate of the Saturday.
        If it is currently Yom Tov, returns the HDate of the first day
        (rather than "leil" Yom Tov). To access Leil Yom Tov, use
        upcoming_shabbat_or_yom_tov.previous_day.
        """
        if self.is_shabbat or self.is_yom_tov:
            return self

        if self.upcoming_yom_tov < self.upcoming_shabbat:
            return self.upcoming_yom_tov
        return self.upcoming_shabbat

    @property
    def first_day(self):
        """Return the first day of Yom Tov or Shabbat.

        This is useful for three-day holidays, for example: it will return the
        first in a string of Yom Tov + Shabbat.
        If this HDate is Shabbat followed by no Yom Tov, returns the Saturday.
        If this HDate is neither Yom Tov, nor Shabbat, this just returns
        itself.
        """
        day_iter = self
        while day_iter.previous_day.is_yom_tov or day_iter.previous_day.is_shabbat:
            day_iter = day_iter.previous_day
        return day_iter

    @property
    def last_day(self):
        """Return the last day of Yom Tov or Shabbat.

        This is useful for three-day holidays, for example: it will return the
        last in a string of Yom Tov + Shabbat.
        If this HDate is Shabbat followed by no Yom Tov, returns the Saturday.
        If this HDate is neither Yom Tov, nor Shabbat, this just returns
        itself.
        """
        day_iter = self
        while day_iter.next_day.is_yom_tov or day_iter.next_day.is_shabbat:
            day_iter = day_iter.next_day
        return day_iter

    def get_holidays_for_year(self, types=None):
        """Get all the actual holiday days for a given HDate's year.

        If specified, use the list of types to limit the holidays returned.
        """
        _LOGGER.debug("Looking up holidays of types %s", types)
        # Filter any non-related holidays depending on Israel/Diaspora only
        holidays_list = [
            holiday
            for holiday in htables.HOLIDAYS
            if (holiday.israel_diaspora == "")
            or (holiday.israel_diaspora == "ISRAEL" and not self.diaspora)
            or (holiday.israel_diaspora == "DIASPORA" and self.diaspora)
        ]

        if types:
            # Filter non-matching holiday types.
            holidays_list = [
                holiday for holiday in holidays_list if holiday.type in types
            ]

        _LOGGER.debug(
            "Holidays after filters have been applied: %s",
            [holiday.name for holiday in holidays_list],
        )

        def holiday_dates_cross_product(holiday):
            """Given a (days, months) pair, compute the cross product.

            If days and/or months are singletons, they are converted to a list.
            """
            return product(
                *([x] if isinstance(x, (int, Months)) else x for x in holiday.date)
            )

        # Compute out every actual Hebrew date on which a holiday falls for
        # this year by exploding out the possible days for each holiday.
        holidays_list = [
            (
                holiday,
                HDate(
                    heb_date=HebrewDate(
                        self.hdate.year, date_instance[1], date_instance[0]
                    ),
                    diaspora=self.diaspora,
                    hebrew=self.hebrew,
                ),
            )
            for holiday in holidays_list
            for date_instance in holiday_dates_cross_product(holiday)
            if len(holiday.date) >= 2
        ]
        # Filter any special cases defined by True/False functions
        holidays_list = [
            (holiday, date)
            for (holiday, date) in holidays_list
            if all(func(date) for func in holiday.date_functions_list)
        ]
        return holidays_list

    @property
    def upcoming_yom_tov(self):
        """Find the next upcoming yom tov (i.e. no-melacha holiday).

        If it is currently the day of yom tov (irrespective of zmanim), returns
        that yom tov.
        """
        if self.is_yom_tov:
            return self
        this_year = self.get_holidays_for_year([HolidayTypes.YOM_TOV])
        next_rosh_hashana = HDate(
            heb_date=HebrewDate(self.hdate.year + 1, Months.TISHREI, 1),
            diaspora=self.diaspora,
            hebrew=self.hebrew,
        )
        next_year = next_rosh_hashana.get_holidays_for_year([HolidayTypes.YOM_TOV])

        # Filter anything that's past.
        holidays_list = [
            holiday_hdate
            for _, holiday_hdate in chain(this_year, next_year)
            if holiday_hdate >= self
        ]

        holidays_list.sort(key=lambda h: h.gdate)

        return holidays_list[0]

    def get_reading(self):
        """Return number of hebrew parasha."""
        _year_type = (self.year_size() % 10) - 3
        year_type = (
            self.diaspora * 1000
            + self.rosh_hashana_dow() * 100
            + _year_type * 10
            + self.pesach_dow()
        )

        _LOGGER.debug("Year type: %d", year_type)

        # Number of days since rosh hashana
        rosh_hashana = HebrewDate(self.hdate.year, Months.TISHREI, 1)
        days = self._jdn - conv.hdate_to_jdn(rosh_hashana)
        # Number of weeks since rosh hashana
        weeks = (days + self.rosh_hashana_dow() - 1) // 7
        _LOGGER.debug("Since Rosh Hashana - Days: %d, Weeks %d", days, weeks)

        # If it's currently Simchat Torah, return VeZot Haberacha.
        if weeks == 3:
            if (
                days <= 22
                and self.diaspora
                and self.dow != 7
                or days <= 21
                and not self.diaspora
            ):
                return 54

        # Special case for Simchat Torah in diaspora.
        if weeks == 4 and days == 22 and self.diaspora:
            return 54

        # Return the indexes for the readings of the given year
        def unpack_readings(readings):
            return list(chain(*([x] if isinstance(x, int) else x for x in readings)))

        reading_for_year = htables.READINGS[year_type]
        readings = unpack_readings(reading_for_year)
        # Maybe recompute the year type based on the upcoming shabbat.
        # This avoids an edge case where today is before Rosh Hashana but
        # Shabbat is in a new year afterwards.
        if (
            weeks >= len(readings)
            and self.hdate.year < self.upcoming_shabbat.hdate.year
        ):
            return self.upcoming_shabbat.get_reading()
        return readings[weeks]


def hebrew_number(num, hebrew=True, short=False):
    """Return "Gimatria" number."""
    if not hebrew:
        return str(num)
    if not 0 <= num < 10000:
        raise ValueError(f"num must be between 0 to 9999, got:{num}")
    hstring = ""
    if num >= 1000:
        hstring += htables.DIGITS[0][num // 1000]
        hstring += "' "
        num = num % 1000
    while num >= 400:
        hstring += htables.DIGITS[2][4]
        num = num - 400
    if num >= 100:
        hstring += htables.DIGITS[2][num // 100]
        num = num % 100
    if num >= 10:
        if num in [15, 16]:
            num = num - 9
        hstring += htables.DIGITS[1][num // 10]
        num = num % 10
    if num > 0:
        hstring += htables.DIGITS[0][num]
    # possibly add the ' and " to hebrew numbers
    if not short:
        if len(hstring) < 2:
            hstring += "'"
        else:
            hstring = hstring[:-1] + '"' + hstring[-1]
    return hstring


def get_omer_string(omer):  # pylint: disable=too-many-branches
    """Return a string representing the count of the Omer."""
    tens = ["", "עשרה", "עשרים", "שלושים", "ארבעים"]
    ones = [
        "",
        "אחד",
        "שנים",
        "שלושה",
        "ארבעה",
        "חמשה",
        "ששה",
        "שבעה",
        "שמונה",
        "תשעה",
    ]
    if not 0 < omer < 50:
        raise ValueError(f"Invalid Omer day: {omer}")
    ten = omer // 10
    one = omer % 10
    omer_string = "היום "
    if 10 < omer < 20:
        omer_string += ones[one] + " עשר"
    elif omer > 9:
        omer_string += ones[one]
        if one:
            omer_string += " ו"
    if omer > 2:
        if omer > 20 or omer in [10, 20]:
            omer_string += tens[ten]
        if omer < 11:
            omer_string += ones[one] + " ימים "
        else:
            omer_string += " יום "
    elif omer == 1:
        omer_string += "יום אחד "
    else:  # omer == 2
        omer_string += "שני ימים "
    if omer > 6:
        omer_string += "שהם "
        weeks = omer // 7
        days = omer % 7
        if weeks > 2:
            omer_string += ones[weeks] + " שבועות "
        elif weeks == 1:
            omer_string += "שבוע אחד "
        else:  # weeks == 2
            omer_string += "שני שבועות "
        if days:
            omer_string += "ו"
            if days > 2:
                omer_string += ones[days] + " ימים "
            elif days == 1:
                omer_string += "יום אחד "
            else:  # days == 2
                omer_string += "שני ימים "
    omer_string += "לעומר"
    return omer_string



"""Location data for HDate - used by Zmanim."""

from dataclasses import dataclass
from datetime import tzinfo

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo


@dataclass
class Location:
    """Define a geolocation for Zmanim calculations."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name="Jerusalem",
        latitude=31.778,
        longitude=35.235,
        timezone="Asia/Jerusalem",
        altitude=754,
        diaspora=False,
    ):
        """Initialitze the location object."""
        self._timezone = None
        self.name = name
        self.latitude = latitude
        self.longitude = longitude
        self.timezone = timezone
        self.altitude = altitude
        self.diaspora = diaspora

    def __repr__(self):
        """Return a representation of Location for programmatic use."""
        return (
            f"Location(name='{self.name}', latitude={self.latitude}, "
            f"longitude={self.longitude}, timezone='{self.timezone}', "
            f"altitude={self.altitude}, diaspora={self.diaspora})"
        )

    @property
    def timezone(self):
        """Return the timezone."""
        return self._timezone

    @timezone.setter
    def timezone(self, value):
        """Set the timezone."""
        self._timezone = value if isinstance(value, tzinfo) else ZoneInfo(value)



"""Tests relating to hebrew numbering with letters."""

import random

import pytest

import hdate.date as dt


class TestOmer:
    """Test get_omer_string"""

    OMER_STRINGS = [
        (1, "היום יום אחד לעומר"),
        (2, "היום שני ימים לעומר"),
        (3, "היום שלושה ימים לעומר"),
        (7, "היום שבעה ימים שהם שבוע אחד לעומר"),
        (8, "היום שמונה ימים שהם שבוע אחד ויום אחד לעומר"),
        (10, "היום עשרה ימים שהם שבוע אחד ושלושה ימים לעומר"),
        (13, "היום שלושה עשר יום שהם שבוע אחד וששה ימים לעומר"),
        (14, "היום ארבעה עשר יום שהם שני שבועות לעומר"),
        (17, "היום שבעה עשר יום שהם שני שבועות ושלושה ימים לעומר"),
        (19, "היום תשעה עשר יום שהם שני שבועות וחמשה ימים לעומר"),
        (28, "היום שמונה ועשרים יום שהם ארבעה שבועות לעומר"),
        (30, "היום שלושים יום שהם ארבעה שבועות ושני ימים לעומר"),
        (37, "היום שבעה ושלושים יום שהם חמשה שבועות ושני ימים לעומר"),
        (45, "היום חמשה וארבעים יום שהם ששה שבועות ושלושה ימים לעומר"),
        (49, "היום תשעה וארבעים יום שהם שבעה שבועות לעומר"),
    ]

    @pytest.mark.parametrize("omer_day,hebrew_string", OMER_STRINGS)
    def test_get_omer_string(self, omer_day, hebrew_string):
        """Test the value returned by calculating the Omer string."""
        assert dt.get_omer_string(omer_day) == hebrew_string

    def test_illegal_value(self):
        """Test passing illegal values to Omer."""
        with pytest.raises(ValueError):
            dt.get_omer_string(random.randint(50, 100))
        with pytest.raises(ValueError):
            dt.get_omer_string(random.randint(-100, 0))


class TestHebrewNumbers:
    """Test hebrew_number"""

    NUMBERS = [
        (1, "א'", "א"),
        (9, "ט'", "ט"),
        (10, "י'", "י"),
        (11, 'י"א', "יא"),
        (15, 'ט"ו', "טו"),
        (127, 'קכ"ז', "קכז"),
        (435, 'תל"ה', "תלה"),
        (770, 'תש"ע', "תשע"),
        (969, 'תתקס"ט', "תתקסט"),
        (1015, "א' ט\"ו", "א' טו"),
    ]

    @pytest.mark.parametrize("number,expected_string,expected_short", NUMBERS)
    def test_hebrew_number(self, number, expected_string, expected_short):
        """Test the calculating the hebrew string."""
        assert dt.hebrew_number(number, short=True) == expected_short
        assert dt.hebrew_number(number) == expected_string

    def test_illegal_value(self):
        """Test unsupported numbers."""
        with pytest.raises(ValueError):
            dt.hebrew_number(random.randint(10000, 20000))
        with pytest.raises(ValueError):
            dt.hebrew_number(random.randint(-100, -1))

    def test_hebrew_number_hebrew_false(self):
        """Test returning a non-hebrew number."""
        number = random.randint(0, 100000)
        assert dt.hebrew_number(number, hebrew=False) == str(number)
        assert dt.hebrew_number(number, hebrew=False, short=True) == str(number)



"""Test HDate objects."""

import datetime
import random

import pytest

import hdate.converters as conv
from hdate import HDate, HebrewDate
from hdate.htables import Months

HEBREW_YEARS_INFO = {
    # year, dow rosh hashana, length, dow pesach
    5753: (2, 353, 3),
    5773: (2, 353, 3),
    5777: (2, 353, 3),
    5756: (2, 355, 5),
    5759: (2, 355, 5),
    5780: (2, 355, 5),
    5783: (2, 355, 5),
    5762: (3, 354, 5),
    5766: (3, 354, 5),
    5769: (3, 354, 5),
    5748: (5, 354, 7),
    5751: (5, 354, 7),
    5758: (5, 354, 7),
    5772: (5, 354, 7),
    5775: (5, 354, 7),
    5778: (5, 354, 7),
    5754: (5, 355, 1),
    5785: (5, 355, 1),
    5761: (7, 353, 1),
    5781: (7, 353, 1),
    5750: (7, 355, 3),
    5764: (7, 355, 3),
    5767: (7, 355, 3),
    5770: (7, 355, 3),
    5788: (7, 355, 3),
    5749: (2, 383, 5),
    5790: (2, 383, 5),
    5752: (2, 385, 7),
    5776: (2, 385, 7),
    5779: (2, 385, 7),
    5755: (3, 384, 7),
    5782: (3, 384, 7),
    5765: (5, 383, 1),
    5768: (5, 383, 1),
    5812: (5, 383, 1),
    5744: (5, 385, 3),
    5771: (5, 385, 3),
    5774: (5, 385, 3),
    5757: (7, 383, 3),
    5784: (7, 383, 3),
    5760: (7, 385, 5),
    5763: (7, 385, 5),
    5787: (7, 385, 5),
}


class TestHDate:
    """Tests for the HDate object."""

    @pytest.fixture
    def default_values(self):
        """Generate an HDate object for today's date."""
        return HDate()

    def test_assign_bad_hdate_value(self):
        """Confirm that bad values raise an error."""
        bad_day_value = HebrewDate(5779, 10, 35)
        with pytest.raises(TypeError):
            HDate().hdate = "not a HebrewDate"
        with pytest.raises(ValueError):
            HebrewDate(5779, 15, 3)
        with pytest.raises(ValueError):
            HDate().hdate = bad_day_value

    @pytest.mark.parametrize("execution_number", list(range(10)))
    def test_random_hdate(self, execution_number, rand_hdate):
        """Run multiple cases with random hdates."""
        print(f"Run number {execution_number}")
        _hdate = HDate()
        _hdate.hdate = rand_hdate.hdate
        assert _hdate.hdate == rand_hdate.hdate
        assert _hdate.gdate == rand_hdate.gdate

    def test_conv_get_size_of_hebrew_year(self):
        """Check that the size of year returned is correct."""
        for year, info in list(HEBREW_YEARS_INFO.items()):
            assert conv.get_size_of_hebrew_year(year) == info[1]

    @pytest.mark.parametrize("execution_number", list(range(10)))
    def test_hdate_get_size_of_hebrew_years(self, execution_number, rand_hdate):
        """Check that the year_size property returns the correct value."""
        print(f"Run number: {execution_number}")
        assert rand_hdate.year_size() == conv.get_size_of_hebrew_year(
            rand_hdate.hdate.year
        )

    def test_rosh_hashana_day_of_week(self, rand_hdate):
        """Check that Rosh Hashana's DOW matches the given dates"""
        for year, info in list(HEBREW_YEARS_INFO.items()):
            rand_hdate.hdate = HebrewDate(
                year, rand_hdate.hdate.month, rand_hdate.hdate.day
            )
            assert rand_hdate.rosh_hashana_dow() == info[0]

    def test_pesach_day_of_week(self, rand_hdate):
        """ "Check tha Pesach DOW matches the given dates."""
        for year, info in list(HEBREW_YEARS_INFO.items()):
            rand_hdate.hdate = HebrewDate(year, 7, 15)
            assert rand_hdate.dow == info[2]
            assert rand_hdate.holiday_name == "pesach"

    UPCOMING_SHABBATOT = [
        ((2018, 11, 30), (2018, 12, 1), (5779, 3, 22)),
        ((2018, 12, 1), (2018, 12, 1), (5779, 3, 23)),
        ((2018, 12, 2), (2018, 12, 8), (5779, 3, 24)),
        ((2018, 12, 3), (2018, 12, 8), (5779, 3, 25)),
        ((2018, 12, 4), (2018, 12, 8), (5779, 3, 26)),
        ((2018, 12, 5), (2018, 12, 8), (5779, 3, 27)),
        ((2018, 12, 6), (2018, 12, 8), (5779, 3, 28)),
        ((2018, 12, 7), (2018, 12, 8), (5779, 3, 29)),
        ((2018, 12, 8), (2018, 12, 8), (5779, 3, 30)),
        ((2018, 12, 9), (2018, 12, 15), (5779, 4, 1)),
    ]

    @pytest.mark.parametrize(
        "current_date, shabbat_date, hebrew_date", UPCOMING_SHABBATOT
    )
    def test_upcoming_shabbat(self, current_date, shabbat_date, hebrew_date):
        """Check the date of the upcoming Shabbat."""
        date = HDate(gdate=datetime.date(*current_date))
        assert date.hdate == HebrewDate(*hebrew_date)
        next_shabbat = date.upcoming_shabbat
        assert next_shabbat.gdate == datetime.date(*shabbat_date)

    def test_prev_and_next_day(self, rand_hdate):
        """Check the previous and next day attributes."""
        assert (rand_hdate.previous_day.gdate - rand_hdate.gdate) == datetime.timedelta(
            -1
        )
        assert (rand_hdate.next_day.gdate - rand_hdate.gdate) == datetime.timedelta(1)


class TestSpecialDays:
    """Test HDate in terms of special days."""

    def test_is_leap_year(self):
        """Test that is_leap_year() working as expected for leap year."""
        leap_date = HDate(heb_date=HebrewDate(5784, 1, 1))
        assert leap_date.is_leap_year

    def test_is_not_leap_year(self):
        """Test that is_leap_year() working as expected for non-leap year."""
        leap_date = HDate(heb_date=HebrewDate(5783, 1, 1))
        assert not leap_date.is_leap_year

    # Test against both a leap year and non-leap year
    @pytest.mark.parametrize(("year"), ((5783, 5784)))
    def test_get_holidays_for_year(self, year):
        """Test that get_holidays_for_year() returns every holiday."""
        cur_date = HDate(heb_date=HebrewDate(year, 1, 1))
        expected_holiday_map = {
            date.gdate: entry for (entry, date) in cur_date.get_holidays_for_year()
        }
        while cur_date.hdate.year == year:
            actual_holiday = cur_date.holiday_name
            if actual_holiday:
                assert actual_holiday == expected_holiday_map[cur_date.gdate].name
            cur_date = cur_date.next_day

    def test_get_holidays_for_year_non_leap_year(self):
        """Test that get_holidays_for_year() returns consistent months."""
        base_date = HDate(heb_date=HebrewDate(5783, 1, 1))
        for _, date in base_date.get_holidays_for_year():
            assert date.hdate.month not in [Months.ADAR_I, Months.ADAR_II]

    def test_get_holidays_for_year_leap_year(self):
        """Test that get_holidays_for_year() returns consistent months."""
        base_date = HDate(heb_date=HebrewDate(5784, 1, 1))
        for _, date in base_date.get_holidays_for_year():
            assert date.hdate.month != Months.ADAR

    NON_MOVING_HOLIDAYS = [
        ((1, 1), "rosh_hashana_i"),
        ((2, 1), "rosh_hashana_ii"),
        ((9, 1), "erev_yom_kippur"),
        ((10, 1), "yom_kippur"),
        ((15, 1), "sukkot"),
        ((17, 1), "hol_hamoed_sukkot"),
        ((18, 1), "hol_hamoed_sukkot"),
        ((19, 1), "hol_hamoed_sukkot"),
        ((20, 1), "hol_hamoed_sukkot"),
        ((21, 1), "hoshana_raba"),
        ((22, 1), "shmini_atzeret"),
        ((15, 7), "pesach"),
        ((17, 7), "hol_hamoed_pesach"),
        ((18, 7), "hol_hamoed_pesach"),
        ((19, 7), "hol_hamoed_pesach"),
        ((20, 7), "hol_hamoed_pesach"),
        ((21, 7), "pesach_vii"),
        ((5, 9), "erev_shavuot"),
        ((6, 9), "shavuot"),
        ((25, 3), "chanukah"),
        ((26, 3), "chanukah"),
        ((27, 3), "chanukah"),
        ((28, 3), "chanukah"),
        ((29, 3), "chanukah"),
        ((1, 4), "chanukah"),
        ((2, 4), "chanukah"),
        ((10, 4), "asara_btevet"),
        ((15, 5), "tu_bshvat"),
        ((18, 8), "lag_bomer"),
        ((15, 11), "tu_bav"),
    ]

    DIASPORA_ISRAEL_HOLIDAYS = [
        # Date, holiday in Diaspora, holiday in Israel
        ((16, 1), "sukkot_ii", "hol_hamoed_sukkot"),
        ((23, 1), "simchat_torah", ""),
        ((16, 7), "pesach_ii", "hol_hamoed_pesach"),
        ((22, 7), "pesach_viii", ""),
        ((7, 9), "shavuot_ii", ""),
    ]

    MOVING_HOLIDAYS = [
        # Possible dates, name
        ([(3, 1), (4, 1)], "tzom_gedaliah"),
        ([(17, 10), (18, 10)], "tzom_tammuz"),
        ([(9, 11), (10, 11)], "tisha_bav"),
    ]

    NEW_HOLIDAYS = [
        # Possible dates, test year range, name
        ([(26, 7), (27, 7), (28, 7)], (5719, 6500), "yom_hashoah"),
        ([(3, 8), (4, 8), (5, 8)], (5709, 5763), "yom_haatzmaut"),
        ([(3, 8), (4, 8), (5, 8), (6, 8)], (5764, 6500), "yom_haatzmaut"),
        ([(2, 8), (3, 8), (4, 8)], (5709, 5763), "yom_hazikaron"),
        ([(2, 8), (3, 8), (4, 8), (5, 8)], (5764, 6500), "yom_hazikaron"),
        ([(28, 8)], (5728, 6500), "yom_yerushalayim"),
        ([(11, 2), (12, 2)], (5758, 6500), "rabin_memorial_day"),
        ([(29, 10)], (5765, 6500), "zeev_zhabotinsky_day"),
        ([(30, 5)], (5734, 6500), "family_day"),
    ]

    ADAR_HOLIDAYS = [
        ([11, 13], "taanit_esther"),
        ([14], "purim"),
        ([15], "shushan_purim"),
        ([7], "memorial_day_unknown"),
    ]

    UPCOMING_HOLIDAYS = [
        ((2018, 8, 8), (2018, 9, 10), "rosh_hashana_i", "BOTH"),
        ((2018, 9, 8), (2018, 9, 10), "rosh_hashana_i", "BOTH"),
        ((2018, 9, 10), (2018, 9, 10), "rosh_hashana_i", "BOTH"),
        ((2018, 9, 11), (2018, 9, 11), "rosh_hashana_ii", "BOTH"),
        ((2018, 9, 12), (2018, 9, 19), "yom_kippur", "BOTH"),
        ((2018, 9, 19), (2018, 9, 19), "yom_kippur", "BOTH"),
        ((2018, 9, 20), (2018, 9, 24), "sukkot", "BOTH"),
        ((2018, 9, 24), (2018, 9, 24), "sukkot", "BOTH"),
        ((2018, 9, 25), (2018, 9, 25), "sukkot_ii", "DIASPORA"),
        ((2018, 9, 25), (2018, 10, 1), "shmini_atzeret", "ISRAEL"),
        ((2018, 9, 26), (2018, 10, 1), "shmini_atzeret", "BOTH"),
        ((2018, 10, 2), (2018, 10, 2), "simchat_torah", "DIASPORA"),
        ((2018, 10, 2), (2019, 4, 20), "pesach", "ISRAEL"),
        ((2018, 10, 3), (2019, 4, 20), "pesach", "BOTH"),
    ]

    @pytest.mark.parametrize(
        "current_date, holiday_date, holiday_name, where", UPCOMING_HOLIDAYS
    )
    def test_get_next_yom_tov(self, current_date, holiday_date, holiday_name, where):
        """Testing the value of next yom tov."""
        print(f"Testing holiday {holiday_name}")
        if where in ("BOTH", "DIASPORA"):
            hdate = HDate(gdate=datetime.date(*current_date), diaspora=True)
            next_yom_tov = hdate.upcoming_yom_tov
            assert next_yom_tov.gdate == datetime.date(*holiday_date)
        if where in ("BOTH", "ISRAEL"):
            hdate = HDate(gdate=datetime.date(*current_date), diaspora=False)
            next_yom_tov = hdate.upcoming_yom_tov
            assert next_yom_tov.gdate == datetime.date(*holiday_date)

    UPCOMING_SHABBAT_OR_YOM_TOV = [
        ((2018, 9, 8), True, {"start": (2018, 9, 8), "end": (2018, 9, 8)}),
        ((2018, 9, 9), True, {"start": (2018, 9, 10), "end": (2018, 9, 11)}),
        ((2018, 9, 16), True, {"start": (2018, 9, 19), "end": (2018, 9, 19)}),
        ((2018, 9, 24), True, {"start": (2018, 9, 24), "end": (2018, 9, 25)}),
        ((2018, 9, 25), True, {"start": (2018, 9, 24), "end": (2018, 9, 25)}),
        ((2018, 9, 24), False, {"start": (2018, 9, 24), "end": (2018, 9, 24)}),
        ((2018, 9, 24), True, {"start": (2018, 9, 24), "end": (2018, 9, 25)}),
        ((2018, 3, 30), True, {"start": (2018, 3, 31), "end": (2018, 4, 1)}),
        ((2018, 3, 30), False, {"start": (2018, 3, 31), "end": (2018, 3, 31)}),
        ((2017, 9, 22), True, {"start": (2017, 9, 21), "end": (2017, 9, 23)}),
        ((2017, 9, 22), False, {"start": (2017, 9, 21), "end": (2017, 9, 23)}),
        ((2017, 10, 4), True, {"start": (2017, 10, 5), "end": (2017, 10, 7)}),
        ((2017, 10, 4), False, {"start": (2017, 10, 5), "end": (2017, 10, 5)}),
        ((2017, 10, 6), True, {"start": (2017, 10, 5), "end": (2017, 10, 7)}),
        ((2017, 10, 6), False, {"start": (2017, 10, 7), "end": (2017, 10, 7)}),
        ((2016, 6, 12), True, {"start": (2016, 6, 11), "end": (2016, 6, 13)}),
    ]

    @pytest.mark.parametrize(
        "current_date, diaspora, dates", UPCOMING_SHABBAT_OR_YOM_TOV
    )
    def test_get_next_shabbat_or_yom_tov(self, current_date, diaspora, dates):
        """Test getting the next shabbat or Yom Tov works."""
        date = HDate(gdate=datetime.date(*current_date), diaspora=diaspora)
        assert date.upcoming_shabbat_or_yom_tov.first_day.gdate == datetime.date(
            *dates["start"]
        )
        assert date.upcoming_shabbat_or_yom_tov.last_day.gdate == datetime.date(
            *dates["end"]
        )

    @pytest.mark.parametrize("date, holiday", NON_MOVING_HOLIDAYS)
    def test_get_holidays_non_moving(self, rand_hdate, date, holiday):
        """Test holidays that have a fixed hebrew date."""
        rand_hdate.hdate = HebrewDate(rand_hdate.hdate.year, date[1], date[0])
        assert rand_hdate.holiday_name == holiday
        assert rand_hdate.is_holiday

    @pytest.mark.parametrize(
        "date, diaspora_holiday, israel_holiday", DIASPORA_ISRAEL_HOLIDAYS
    )
    def test_get_diaspora_israel_holidays(
        self, rand_hdate, date, diaspora_holiday, israel_holiday
    ):
        """Test holidays that differ based on diaspora/israel."""
        rand_hdate.hdate = HebrewDate(rand_hdate.hdate.year, date[1], date[0])
        assert rand_hdate.holiday_name == israel_holiday
        rand_hdate.diaspora = True
        assert rand_hdate.holiday_name == diaspora_holiday
        assert rand_hdate.is_holiday

    @pytest.mark.parametrize("possible_dates, holiday", MOVING_HOLIDAYS)
    def test_get_holidays_moving(self, possible_dates, holiday):
        """Test holidays that are moved based on the DOW."""
        found_matching_holiday = False
        year = random.randint(5000, 6500)

        print("Testing " + holiday + " for " + str(year))

        for date in possible_dates:
            date_under_test = HDate(hebrew=False)
            date_under_test.hdate = HebrewDate(year, date[1], date[0])
            if date_under_test.holiday_name == holiday:
                print("date ", date_under_test, " matched")
                for other in possible_dates:
                    if other != date:
                        other_date = HDate(hebrew=False)
                        other_date.hdate = HebrewDate(year, other[1], other[0])
                        print("checking ", other_date, " doesn't match")
                        assert other_date.holiday_name != holiday
                found_matching_holiday = True
                assert date_under_test.is_holiday

        assert found_matching_holiday

    @pytest.mark.parametrize("possible_dates, years, holiday", NEW_HOLIDAYS)
    def test_new_holidays_multiple_date(self, possible_dates, years, holiday):
        """Test holidays that have multiple possible dates."""
        found_matching_holiday = False
        year = random.randint(*years)

        print(f"Testing {holiday} for {year}")

        for date in possible_dates:
            date_under_test = HDate(hebrew=False)
            date_under_test.hdate = HebrewDate(year, date[1], date[0])
            if date_under_test.holiday_name == holiday:
                print("date ", date_under_test, " matched")
                for other in possible_dates:
                    if other != date:
                        other_date = HDate(hebrew=False)
                        other_date.hdate = HebrewDate(year, other[1], other[0])
                        print("checking ", other_date, " doesn't match")
                        assert other_date.holiday_name != holiday
                found_matching_holiday = True
                assert date_under_test.is_holiday

        assert found_matching_holiday

    @pytest.mark.parametrize("possible_dates, years, holiday", NEW_HOLIDAYS)
    def test_new_holidays_invalid_before(self, possible_dates, years, holiday):
        """Test holidays that were created over time."""
        # Yom hazikaron and yom ha'atsmaut don't test for before 5764
        if years[0] == 5764 and holiday in ["yom_hazikaron", "yom_haatzmaut"]:
            return
        year = random.randint(5000, years[0] - 1)
        print("Testing " + holiday + " for " + str(year))
        for date in possible_dates:
            date_under_test = HDate()
            date_under_test.hdate = HebrewDate(year, date[1], date[0])
            assert date_under_test.holiday_name == ""

    def test_get_holiday_hanuka_3rd_tevet(self):
        """Test Chanuka falling on 3rd of Tevet."""
        year = random.randint(5000, 6000)
        year_size = conv.get_size_of_hebrew_year(year)
        myhdate = HDate(heb_date=HebrewDate(year, 4, 3))
        print(year_size)
        if year_size in [353, 383]:
            assert myhdate.holiday_name == "chanukah"
        else:
            assert myhdate.holiday_name == ""

    @pytest.mark.parametrize("possible_days, holiday", ADAR_HOLIDAYS)
    def test_get_holiday_adar(self, possible_days, holiday):
        """Test holidays for Adar I/Adar II."""
        year = random.randint(5000, 6000)
        year_size = conv.get_size_of_hebrew_year(year)
        month = 6 if year_size < 360 else 14
        myhdate = HDate()

        for day in possible_days:
            myhdate.hdate = HebrewDate(year, month, day)
            if day == 13 and myhdate.dow == 7 and holiday == "taanit_esther":
                assert myhdate.holiday_name == ""
            elif day == 11 and myhdate.dow != 5 and holiday == "taanit_esther":
                assert myhdate.holiday_name == ""
            else:
                assert myhdate.holiday_name == holiday

    @pytest.mark.parametrize("execution_number", list(range(10)))
    def test_get_omer_day(self, execution_number, rand_hdate):
        """Test value of the Omer."""
        print(f"Test number {execution_number}")
        if (
            rand_hdate.hdate.month not in [Months.NISAN, Months.IYYAR, Months.SIVAN]
            or rand_hdate.hdate.month == Months.NISAN
            and rand_hdate.hdate.day < 16
            or rand_hdate.hdate.month == Months.SIVAN
            and rand_hdate.hdate.day > 5
        ):
            assert rand_hdate.omer_day == 0

        nissan = list(range(16, 30))
        iyyar = list(range(1, 29))
        sivan = list(range(1, 5))

        for day in nissan:
            rand_hdate.hdate = HebrewDate(rand_hdate.hdate.year, 7, day)
            assert rand_hdate.omer_day == day - 15
        for day in iyyar:
            rand_hdate.hdate = HebrewDate(rand_hdate.hdate.year, 8, day)
            assert rand_hdate.omer_day == day + 15
        for day in sivan:
            rand_hdate.hdate = HebrewDate(rand_hdate.hdate.year, 9, day)
            assert rand_hdate.omer_day == day + 44

    def test_daf_yomi(self):
        """Test value of Daf Yomi."""
        # Random test date
        myhdate = HDate(gdate=datetime.date(2014, 4, 28), hebrew=False)
        assert myhdate.daf_yomi == "Beitzah 29"
        # Beginning/end of cycle:
        myhdate = HDate(gdate=datetime.date(2020, 1, 4), hebrew=False)
        assert myhdate.daf_yomi == "Niddah 73"
        myhdate = HDate(gdate=datetime.date(2020, 1, 5), hebrew=False)
        assert myhdate.daf_yomi == "Berachos 2"
        myhdate = HDate(gdate=datetime.date(2020, 3, 7), hebrew=True)
        assert myhdate.daf_yomi == "ברכות סד"
        myhdate = HDate(gdate=datetime.date(2020, 3, 8), hebrew=True)
        assert myhdate.daf_yomi == "שבת ב"


class TestHDateReading:
    """Tests dealing with parshat hashavua."""

    READINGS_FOR_YEAR_DIASPORA = [
        # שנים מעוברות
        # זשה
        (
            5763,
            [
                [0, 53, 0],
                list(range(29)),
                [0],
                list(range(29, 35)),
                [0],
                list(range(35, 39)),
                [59, 41, 60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זחג
        (
            5757,
            [
                [0, 53, 0],
                list(range(29)),
                [0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # השג
        (5774, [[53, 0], list(range(30)), [0], list(range(30, 51)), [61]]),
        # החא
        (5768, [[53, 0], list(range(30)), [0], list(range(30, 51))]),
        # גכז
        (
            5755,
            [
                [52, 53],
                list(range(29)),
                [0, 0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
            ],
        ),
        # בשז
        (
            5776,
            [
                [52, 53],
                list(range(29)),
                [0, 0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
            ],
        ),
        # בחה
        (
            5749,
            [
                [52, 53],
                list(range(29)),
                [0],
                list(range(29, 35)),
                [0],
                list(range(35, 39)),
                [59, 41, 60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # שנים פשוטות
        # השא
        (
            5754,
            [
                [53, 0],
                list(range(26)),
                [0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 54)),
            ],
        ),
        # בשה
        (
            5756,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58, 34, 0],
                list(range(35, 39)),
                [59, 41, 60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זחא
        (
            5761,
            [
                [0, 53, 0],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 52)),
            ],
        ),
        # גכה
        (
            5769,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58, 34, 0],
                list(range(35, 39)),
                [59, 41, 60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זשג
        (
            5770,
            [
                [0, 53, 0],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # הכז
        (
            5775,
            [
                [53, 0],
                list(range(22)),
                [55, 24, 25, 0, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
            ],
        ),
        # בחג
        (
            5777,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        (
            5778,
            [
                [53, 0],
                list(range(22)),
                [55, 24, 25, 0, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 52)),
            ],
        ),
    ]

    READINGS_FOR_YEAR_ISRAEL = [
        # שנים מעוברות
        # זשה
        (
            5763,
            [
                [0, 53, 0, 54],
                list(range(1, 29)),
                [0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זחג
        (
            5757,
            [
                [0, 53, 0, 54],
                list(range(1, 29)),
                [0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # השג
        (5774, [[53, 0], list(range(30)), [0], list(range(30, 51)), [61]]),
        # החא
        (5768, [[53, 0], list(range(30)), [0], list(range(30, 51))]),
        # גכז
        (5755, [[52, 53], list(range(29)), [0], list(range(29, 51))]),
        # בשז
        (5776, [[52, 53], list(range(29)), [0], list(range(29, 51))]),
        # בחה
        (
            5749,
            [
                [52, 53],
                list(range(29)),
                [0],
                list(range(29, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # שנים פשוטות
        # השא
        (
            5754,
            [
                [53, 0],
                list(range(26)),
                [0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 54)),
            ],
        ),
        # בשה
        (
            5756,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זחא
        (
            5761,
            [
                [0, 53, 0, 54],
                list(range(1, 22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 52)),
            ],
        ),
        # גכה
        (
            5769,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # זשג
        (
            5770,
            [
                [0, 53, 0, 54],
                list(range(1, 22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
        # הכז
        (
            5775,
            [
                [53, 0],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57],
                list(range(31, 42)),
                [60],
                list(range(44, 52)),
            ],
        ),
        # בחג
        (
            5777,
            [
                [52, 53],
                list(range(22)),
                [55, 24, 25, 0, 26, 56, 57, 31, 58],
                list(range(34, 42)),
                [60],
                list(range(44, 51)),
                [61],
            ],
        ),
    ]

    @pytest.mark.parametrize("year, parshiyot", READINGS_FOR_YEAR_ISRAEL)
    def test_get_reading_israel(self, year, parshiyot):
        """Test parshat hashavua in Israel."""
        mydate = HDate(hebrew=False, diaspora=False)
        mydate.hdate = HebrewDate(year, 1, 1)

        # Get next Saturday
        tdelta = datetime.timedelta((12 - mydate.gdate.weekday()) % 7)
        mydate.gdate += tdelta

        shabatot = [item for subl in parshiyot for item in subl]
        for shabbat in shabatot:
            print("Testing: ", mydate)
            assert mydate.get_reading() == shabbat
            mydate.gdate += datetime.timedelta(days=7)
        mydate.hdate = HebrewDate(year, 1, 22)
        # VeZot Habracha in Israel always falls on 22 of Tishri
        assert mydate.get_reading() == 54

    @pytest.mark.parametrize("year, parshiyot", READINGS_FOR_YEAR_DIASPORA)
    def test_get_reading_diaspora(self, year, parshiyot):
        """Test parshat hashavua in the diaspora."""
        mydate = HDate(hebrew=False, diaspora=True)
        mydate.hdate = HebrewDate(year, 1, 1)

        # Get next Saturday
        tdelta = datetime.timedelta((12 - mydate.gdate.weekday()) % 7)
        mydate.gdate += tdelta

        shabatot = [item for subl in parshiyot for item in subl]
        for shabbat in shabatot:
            print("Testing: ", mydate)
            assert mydate.get_reading() == shabbat
            mydate.gdate += datetime.timedelta(days=7)
        mydate.hdate = HebrewDate(year, 1, 23)
        # VeZot Habracha in Israel always falls on 22 of Tishri
        assert mydate.get_reading() == 54

    @pytest.mark.parametrize("year", range(5740, 5800))
    def test_nitzavim_always_before_rosh_hashana(self, year):
        """A property: Nitzavim alway falls before rosh hashana."""
        mydate = HDate(hebrew=False, diaspora=False)
        mydate.hdate = HebrewDate(year, Months.TISHREI, 1)
        tdelta = datetime.timedelta((12 - mydate.gdate.weekday()) % 7 - 7)
        # Go back to the previous shabbat
        mydate.gdate += tdelta
        print("Testing date: {mydate} which is {tdelta} days before Rosh Hashana")
        assert mydate.get_reading() in [51, 61]

    @pytest.mark.parametrize("year", range(5740, 5800))
    def test_vayelech_or_haazinu_always_after_rosh_hashana(self, year):
        """A property: Vayelech or Haazinu always falls after rosh hashana."""
        mydate = HDate(hebrew=False, diaspora=True)
        mydate.hdate = HebrewDate(year, Months.TISHREI, 1)
        tdelta = datetime.timedelta((12 - mydate.gdate.weekday()) % 7)
        # Go to the next shabbat (unless shabbat falls on Rosh Hashana)
        mydate.gdate += tdelta
        print(f"Testing date: {mydate} which is {tdelta} days after Rosh Hashana")
        assert mydate.get_reading() in [52, 53, 0]

    def test_last_week_of_the_year(self):
        """The last day of the year is parshat Vayelech."""
        mydate = HDate()
        mydate.hdate = HebrewDate(5779, Months.ELUL, 29)
        assert mydate.get_reading() == 52



"""Methods for going back and forth between various calendars."""

import datetime

from hdate.hebrew_date import HebrewDate
from hdate.htables import Months


def get_chalakim(hours, parts):
    """Return the number of total parts (chalakim)."""
    return (hours * PARTS_IN_HOUR) + parts


PARTS_IN_HOUR = 1080
PARTS_IN_DAY = 24 * PARTS_IN_HOUR
PARTS_IN_WEEK = 7 * PARTS_IN_DAY
PARTS_IN_MONTH = PARTS_IN_DAY + get_chalakim(12, 793)  # Fix for regular month


def _days_from_3744(hebrew_year):
    """Return: Number of days since 3,1,3744."""
    # Start point for calculation is Molad new year 3744 (16BC)
    years_from_3744 = hebrew_year - 3744
    molad_3744 = get_chalakim(1 + 6, 779)  # Molad 3744 + 6 hours in parts

    # Time in months

    # Number of leap months
    leap_months = (years_from_3744 * 7 + 1) // 19
    leap_left = (years_from_3744 * 7 + 1) % 19  # Months left of leap cycle
    months = years_from_3744 * 12 + leap_months  # Total Number of months

    # Time in parts and days
    # Molad This year + Molad 3744 - corrections
    parts = months * PARTS_IN_MONTH + molad_3744
    # 28 days in month + corrections
    days = months * 28 + parts // PARTS_IN_DAY - 2

    # Time left for round date in corrections
    # 28 % 7 = 0 so only corrections counts
    parts_left_in_week = parts % PARTS_IN_WEEK
    parts_left_in_day = parts % PARTS_IN_DAY
    week_day = parts_left_in_week // PARTS_IN_DAY

    # pylint: disable=too-many-boolean-expressions
    # pylint-comment: Splitting the 'if' below might create a bug in case
    # the order is not kept.

    # Molad ד"ר ט"ג
    if (
        (
            leap_left < 12
            and week_day == 3
            and parts_left_in_day >= get_chalakim(9 + 6, 204)
        )
        or
        # Molad ט"פקת ו"טב
        (
            leap_left < 7
            and week_day == 2
            and parts_left_in_day >= get_chalakim(15 + 6, 589)
        )
    ):
        days += 1
        week_day += 1

    # pylint: enable=too-many-boolean-expressions

    # ADU
    if week_day in (1, 4, 6):
        days += 1

    return days


def get_size_of_hebrew_year(hebrew_year):
    """Return: total days in hebrew year."""
    return _days_from_3744(hebrew_year + 1) - _days_from_3744(hebrew_year)


def gdate_to_jdn(date):
    """
    Compute Julian day from Gregorian day, month and year.

    Algorithm from wikipedia's julian_day article.
    Return: The julian day number
    """
    not_jan_or_feb = (14 - date.month) // 12
    year_since_4800bc = date.year + 4800 - not_jan_or_feb
    month_since_4800bc = date.month + 12 * not_jan_or_feb - 3
    jdn = (
        date.day
        + (153 * month_since_4800bc + 2) // 5
        + 365 * year_since_4800bc
        + (year_since_4800bc // 4 - year_since_4800bc // 100 + year_since_4800bc // 400)
        - 32045
    )
    return jdn


def hdate_to_jdn(date):
    """
    Compute Julian day from Hebrew day, month and year.

    Return: julian day number,
            1 of tishrey julians,
            1 of tishrey julians next year
    """
    day = date.day
    month = date.month.value
    if date.month == Months.ADAR_I:
        month = 6
    if date.month == Months.ADAR_II:
        month = 6
        day += 30

    # Calculate days since 1,1,3744
    day = _days_from_3744(date.year) + (59 * (month - 1) + 1) // 2 + day

    # length of year
    length_of_year = get_size_of_hebrew_year(date.year)
    # Special cases for this year
    if length_of_year % 10 > 4 and month > 2:  # long Heshvan
        day += 1
    if length_of_year % 10 < 4 and month > 3:  # short Kislev
        day -= 1
    if length_of_year > 365 and month > 6:  # leap year
        day += 30

    # adjust to julian
    return day + 1715118


def jdn_to_gdate(jdn):
    """
    Convert from the Julian day to the Gregorian day.

    Algorithm from 'Julian and Gregorian Day Numbers' by Peter Meyer.
    Return: day, month, year
    """
    # pylint: disable=invalid-name

    # The algorithm is a verbatim copy from Peter Meyer's article
    # No explanation in the article is given for the variables
    # Hence the exceptions for pylint and for flake8 (E741)

    l = jdn + 68569  # noqa: E741
    n = (4 * l) // 146097
    l = l - (146097 * n + 3) // 4  # noqa: E741
    i = (4000 * (l + 1)) // 1461001  # that's 1,461,001
    l = l - (1461 * i) // 4 + 31  # noqa: E741
    j = (80 * l) // 2447
    day = l - (2447 * j) // 80
    l = j // 11  # noqa: E741
    month = j + 2 - (12 * l)
    year = 100 * (n - 49) + i + l  # that's a lower-case L

    return datetime.date(year, month, day)


def jdn_to_hdate(jdn):
    """Convert from the Julian day to the Hebrew day."""
    # calculate Gregorian date
    date = jdn_to_gdate(jdn)

    # Guess Hebrew year is Gregorian year + 3760
    year = date.year + 3760

    jdn_tishrey1 = hdate_to_jdn(HebrewDate(year, Months.TISHREI, 1))
    jdn_tishrey1_next_year = hdate_to_jdn(HebrewDate(year + 1, Months.TISHREI, 1))

    # Check if computed year was underestimated
    if jdn_tishrey1_next_year <= jdn:
        year = year + 1
        jdn_tishrey1 = jdn_tishrey1_next_year
        jdn_tishrey1_next_year = hdate_to_jdn(HebrewDate(year + 1, Months.TISHREI, 1))

    size_of_year = get_size_of_hebrew_year(year)

    # days into this year, first month 0..29
    days = jdn - jdn_tishrey1

    # last 8 months always have 236 days
    if days >= (size_of_year - 236):  # in last 8 months
        days = days - (size_of_year - 236)
        month = days * 2 // 59
        day = days - (month * 59 + 1) // 2 + 1

        month = month + 4 + 1

        # if leap
        if size_of_year > 355 and month <= 6:
            month = month + 8
    else:  # in 4-5 first months
        # Special cases for this year
        if size_of_year % 10 > 4 and days == 59:  # long Heshvan (day 30)
            month = 1
            day = 30
        elif size_of_year % 10 > 4 and days > 59:  # long Heshvan
            month = (days - 1) * 2 // 59
            day = days - (month * 59 + 1) // 2
        elif size_of_year % 10 < 4 and days > 87:  # short kislev
            month = (days + 1) * 2 // 59
            day = days - (month * 59 + 1) // 2 + 2
        else:  # regular months
            month = days * 2 // 59
            day = days - (month * 59 + 1) // 2 + 1

        month = month + 1

    return HebrewDate(year, Months(month), day)
